import numpy as np

from holisticai.utils.transformers.bias import SensitiveGroups


class WassersteinBarycenterAlgorithm:
    def __init__(self):
        self.sens_groups = SensitiveGroups()
        self.eps = np.finfo(float).eps

    def fit(self, y_pred: np.ndarray, sensitive_groups: np.ndarray):

        p_attr = self.sens_groups.fit_transform(
            sensitive_groups, convert_numeric=True
        ).squeeze()
        self.group_values = np.unique(p_attr)

        group_freq = [
            np.sum(p_attr == self.group_values[0]),
            np.sum(p_attr == self.group_values[1]),
        ]

        self.iM = np.argmax(group_freq)
        self.nM = group_freq[self.iM]

        self.im = np.argmin(group_freq)
        self.nm = group_freq[self.im]

        self.p = self.nm / len(p_attr)
        self.q = 1 - self.p

        self.SL = p_attr

        noise = self.eps * np.random.randn(len(y_pred)).squeeze()
        self.YL = y_pred + noise

        self.minY = np.min(self.YL)
        self.maxY = np.max(self.YL)

    def _optimize_ts(self, yt, i1, i2, n1, n2):
        dist_best = np.inf
        for t in np.linspace(self.minY, self.maxY, 100):
            tmp1 = np.sum(self.YL[self.SL == self.group_values[i1]] < t) / n1
            tmp2 = np.sum(self.YL[self.SL == self.group_values[i2]] < yt) / n2
            dist = np.abs(tmp1 - tmp2)
            if dist_best > dist:
                dist_best = dist
                ts = t
        return ts

    def _update_yt(self, yt, group):
        if group == self.group_values[self.im]:
            ts = self._optimize_ts(yt, self.iM, self.im, self.nM, self.nm)
            yf = self.p * yt + (1 - self.p) * ts

        else:
            ts = self._optimize_ts(yt, self.im, self.iM, self.nm, self.nM)
            yf = self.q * yt + (1 - self.q) * ts
        return yf

    def transform(self, y_pred: np.ndarray, sensitive_groups: np.ndarray):
        ST = self.sens_groups.transform(sensitive_groups, convert_numeric=True)
        noise = self.eps * np.random.randn(len(y_pred)).squeeze()
        YT = y_pred + noise
        YF = np.array([self._update_yt(yt, st) for yt, st in zip(YT, ST)])
        return YF



import matplotlib.pyplot as plt
import numpy as np


def distance(a, b, order=2):
    """
    Calculates the specified norm between two vectors.

    Args:
            a (list) : First vector
            b (list) : Second vector
            order (int) : Order of the norm to be calculated as distance

    Returns:
            Resultant norm value
    """
    assert len(a) == len(b), "Length of the vectors for distance don't match."
    return np.linalg.norm(x=np.array(a) - np.array(b), ord=order)


def balance_calculation(data, centers, mapping):
    """
    Checks fairness for each of the clusters defined by k-centers.
    Returns balance using the total and class counts.

    Args:
            data (list)
            centers (list)
            mapping (list) : tuples of the form (data, center)

    Returns:
            fair (dict) : key=center, value=(sum of 1's corresponding to fairness variable, number of points in center)
    """
    fair = dict([(i, [0, 0]) for i in centers])
    for i in mapping:
        fair[i[1]][1] += 1
        if data[i[0]][0] == 1:  # MARITAL
            fair[i[1]][0] += 1

    curr_b = []
    for i in list(fair.keys()):
        p = fair[i][0]
        q = fair[i][1] - fair[i][0]
        if p == 0 or q == 0:
            balance = 0
        else:
            balance = min(float(p / q), float(q / p))
        curr_b.append(balance)

    return min(curr_b)


def plot_analysis(degrees, costs, balances, step_size):
    """
    Plots the curves for costs and balances.

    Args:
            degrees (list)
            costs (list)
            balances (list)
            step_size (int)
    """
    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(14, 5))
    ax[0].plot(costs, marker=".", color="blue")
    ax[0].set_xticks(list(range(0, len(degrees), step_size)))
    ax[0].set_xticklabels(
        list(range(min(degrees), max(degrees) + 1, step_size)), fontsize=12
    )
    ax[1].plot(balances, marker="x", color="saddlebrown")
    ax[1].set_xticks(list(range(0, len(degrees), step_size)))
    ax[1].set_xticklabels(
        list(range(min(degrees), max(degrees) + 1, step_size)), fontsize=12
    )
    plt.show()



import numpy as np
from sklearn.metrics import pairwise_distances_chunked as pdist_chunk

from ._bound_update import get_S_discrete
from ._method_utils import (
    KernelBound_k,
    NormalizedCutEnergy,
    NormalizedCutEnergy_discrete,
    ecdist,
    km_discrete_energy,
    km_le,
    reduce_func,
)
from ._utils import create_affinity


class KUtils:
    def __init__(self, X, K):
        self.N = len(X)
        self.K = K
        self.X = X

    def update(self, a_p, l, C):
        S = get_S_discrete(l, self.N, self.K)
        l = km_le(self.X, C)
        return S, l

    def compute_clustering_energy(self, C, l, S):
        """
        compute fair clustering energy
        """
        e_dist = self.compute_a_p(l, C)
        clustering_E = (S * e_dist).sum()
        clustering_E_discrete = [
            km_discrete_energy(e_dist, l, k) for k in range(self.K)
        ]
        clustering_E_discrete = sum(clustering_E_discrete)
        return clustering_E, clustering_E_discrete


class KmeansUtils(KUtils):
    def compute_a_p(self, l, C):
        a_p = ecdist(self.X, C, squared=True)
        return a_p

    def step(self, l, C):
        tmp_list = [np.where(l == k)[0] for k in range(self.K)]
        C_list = [self.kmeans_update(t) for t in tmp_list]
        C = np.asarray(np.vstack(C_list))
        sqdist = ecdist(self.X, C, squared=True)
        a_p = sqdist.copy()
        return a_p, C

    def kmeans_update(self, tmp):
        """ """
        X_tmp = self.X[tmp, :]
        c1 = X_tmp.mean(axis=0)

        return c1


class KmedianUtils(KUtils):
    def compute_a_p(self, l, C):
        a_p = ecdist(self.X, C)
        return a_p

    def step(self, l, C):
        tmp_list = [np.where(l == k)[0] for k in range(self.K)]
        C_list = [self.kmedian_update(x) for x in tmp_list]
        C = np.asarray(np.vstack(C_list))
        sqdist = ecdist(self.X, C)
        a_p = sqdist.copy()
        return a_p, C

    def kmedian_update(self, tmp):
        # print("ID of process running worker: {}".format(os.getpid()))
        X_tmp = self.X[tmp, :]
        D = pdist_chunk(X_tmp, reduce_func=reduce_func)
        J = next(D)
        j = np.argmin(J)
        c1 = X_tmp[j, :]
        return c1


class NCutUtils:
    def __init__(self, X, K):
        self.K = K
        self.N = len(X)
        self.X = X
        self.knn = 20

    def compute_a_p(self, l, C):
        self.A = create_affinity(self.X, self.knn)
        self.d = self.A.sum(axis=1)
        S = get_S_discrete(l, self.N, self.K)
        sqdist_list = [
            KernelBound_k(self.A, self.d, S[:, k], self.N) for k in range(self.K)
        ]
        sqdist = np.asarray(np.vstack(sqdist_list).T)
        a_p = sqdist.copy()
        return a_p, C

    def step(self, l, C):
        S = get_S_discrete(l, self.N, self.K)
        sqdist_list = [
            KernelBound_k(self.A, self.d, S[:, k], self.N) for k in range(self.K)
        ]
        sqdist = np.asarray(np.vstack(sqdist_list).T)
        a_p = sqdist.copy()
        return a_p

    def update(self, a_p, l, C):
        l = a_p.argmin(axis=1)
        S = get_S_discrete(l, self.N, self.K)
        return S, l

    def compute_clustering_energy(self, C, l, S):
        """
        compute fair clustering energy
        """
        clustering_E = NormalizedCutEnergy(self.A, S, l)
        clustering_E_discrete = NormalizedCutEnergy_discrete(self.A, l)
        return clustering_E, clustering_E_discrete



import numpy as np

from ._utils import distance


class KCenters(object):
    def __init__(self, n_clusters=5):
        """
        k (int) : Number of centers to be identified
        """
        self.k = n_clusters

    def fit(self, data):
        """
        Performs the k-centers algorithm.

        Args:
                data (list) : Points in the dataset
        """
        # Randomly choosing an initial center
        # random.seed(42)

        self.data = data
        self.centers = [int(np.random.randint(0, len(self.data), 1))]
        self.costs = []

        while True:
            # Remaining points in the data set
            rem_points = list(set(range(0, len(self.data))) - set(self.centers))

            # Finding the point which has the closest center most far-off
            point_center = [
                (i, min([distance(self.data[i], self.data[j]) for j in self.centers]))
                for i in rem_points
            ]
            point_center = sorted(point_center, key=lambda x: x[1], reverse=True)

            self.costs.append(point_center[0][1])
            if len(self.centers) < self.k:
                self.centers.append(point_center[0][0])
            else:
                break

        self.cluster_centers_ = [self.data[j] for j in self.centers]
        self.labels = self.assign()
        return

    def assign(self):
        """
        Assigning every point in the dataset to the closest center.

        Returns:
                mapping (list) : tuples of the form (point, center)
        """
        mapping = [
            (
                i,
                sorted(
                    [(j, distance(self.data[i], self.data[j])) for j in self.centers],
                    key=lambda x: x[1],
                    reverse=False,
                )[0][0],
            )
            for i in range(len(self.data))
        ]

        return mapping



import math
from collections import defaultdict

import numpy as np
from sklearn.cluster import KMeans

from ._bound_update import BoundUpdate, normalize_2
from ._logger import MFLogger
from ._method_utils import _init_centroids, compute_fairness_energy, ecdist, km_le
from ._methods import KmeansUtils, KmedianUtils, NCutUtils
from ._utils import get_fair_accuracy_proportional


class Monitor:
    def __init__(self):
        self.history = defaultdict(list)

    def save(self, **kargs):
        for k, v in kargs.items():
            self.history[k].append(v)


def km_init(K, X, C_init, l_init=None, random_state=np.random.RandomState()):
    """
    Initial seeds
    """
    if isinstance(C_init, str):

        if C_init == "kmeans_plus":
            M = _init_centroids(
                X, n_clusters=K, init="k-means++", random_state=random_state
            )
            l = km_le(X, M)

        elif C_init == "kmeans":
            kmeans = KMeans(n_clusters=K).fit(X)
            l = kmeans.labels_
            M = kmeans.cluster_centers_
    else:
        M = C_init.copy()
        # l = km_le(X,M)
        l = l_init.copy()

    del C_init, l_init

    return M, l


class FairnessUtility:
    def restore_nonempty_cluster(self, X, K, oldl, oldC, oldS):
        ts_limit = 2
        C_init = "kmeans"
        if self.ts > ts_limit:
            print("not having some labels")
            trivial_status = True
            l = oldl.copy()
            C = oldC.copy()
            S = oldS.copy()

        else:
            print("try with new seeds")

            C, l = km_init(self.K, X, C_init)
            sqdist = ecdist(X, C, squared=True)
            S = normalize_2(np.exp((-sqdist)))
            trivial_status = False

        return l, C, S, trivial_status

    def fit_fairness(self, X, a_p, group_prob, groups_ids):
        trivial_status = False
        l_check = a_p.argmin(axis=1)

        # Check for empty cluster
        if len(np.unique(l_check)) != self.K:
            l, C, S, trivial_status = self.restore_nonempty_cluster(
                X, self.K, self.old["l"], self.old["C"], self.old["S"]
            )
            self.ts = self.ts + 1
            return S, l, trivial_status

        l, S, bound_E = self.bound_update.transform(a_p, group_prob, groups_ids)
        return S, l, trivial_status


METHODS_CATALOG = {"kmeans": KmeansUtils, "kmedian": KmedianUtils, "ncut": NCutUtils}


class FairClustering(FairnessUtility):
    def __init__(self, K, L, lmbda, method, max_iter=100, verbose=True):
        self.K = K
        self.lmbda = lmbda
        self.L = L
        self.bound_update = BoundUpdate(bound_lambda=lmbda, L=L, bound_iteration=10000)
        self.method = method
        self.monitor = Monitor()
        self.max_iter = max_iter
        self.logger = MFLogger(total_iterations=self.max_iter, verbose=verbose)

    def fit(self, X, group_prob, groups_ids, random_state):

        C, l = km_init(self.K, X, "kmeans_plus", random_state=random_state)

        self.clustering_method = METHODS_CATALOG[self.method](X, self.K)
        self.ts = 0
        S = []
        fairness_error = 0.0
        oldE = 1e100
        for i in range(self.max_iter):
            self.old = {"C": C.copy(), "l": l.copy(), "S": S.copy()}

            if i == 0:
                a_p = self.clustering_method.compute_a_p(l, C)
            else:
                a_p, C = self.clustering_method.step(l, C)

            S, l, trivial_status = self.fit_fairness(X, a_p, group_prob, groups_ids)
            if trivial_status:
                break

            (
                clusterE,
                clusterE_discrete,
            ) = self.clustering_method.compute_clustering_energy(C, l, S)
            fairness_error = get_fair_accuracy_proportional(
                group_prob, groups_ids, l, self.K
            )
            fairE = compute_fairness_energy(group_prob, groups_ids, S, self.lmbda)
            currentE = clusterE + fairE

            self.monitor.save(
                fair_cluster_E=currentE,
                cluster_E=clusterE,
                fair_E=fairE,
                cluster_E_discrete=clusterE_discrete,
                fairness_error=fairness_error,
            )

            self.logger.update(
                iteration=i,
                fairness_error=fairness_error,
                fair_cluster_energy=currentE,
                cluster_energy=clusterE_discrete,
            )

            if (len(np.unique(l)) != self.K) or math.isnan(fairness_error):
                l, C, S, trivial_status = self.restore_nonempty_cluster(
                    X, self.K, self.old["l"], self.old["C"], self.old["S"]
                )
                self.ts = self.ts + 1
                if trivial_status:
                    break

            if i > 1 and (abs(currentE - oldE) <= 1e-4 * abs(oldE)):
                break

            else:
                oldE = currentE.copy()

        self.C = C
        self.l = l
        self.S = S
        self.group_prob = group_prob
        return self

    def predict(self, X, groups_ids):
        S = ecdist(X, self.C, squared=True)
        l = np.argmax(S, axis=1)
        C = self.C
        N, D = X.shape
        group_prob = self.group_prob

        self.clustering_method = METHODS_CATALOG[self.method](X, self.K)
        a_p = self.clustering_method.compute_a_p(l, C)
        _, l, _ = self.fit_fairness(X, a_p, group_prob, groups_ids)
        return l



# Base Imports
import numpy as np
import pandas as pd

# utils
from ...utils._validation import _multiclass_checks


def confusion_matrix(y_pred, y_true, classes=None, normalize=None):
    """
    Confusion Matrix.

    Description
    ----------
    This function computes the confusion matrix. The i,jth
    entry is the number of elements with predicted class i
    and true class j.

    Parameters
    ----------
    y_pred : array-like
        Prediction vector (categorical)
    y_true : array-like
        Target vector (categorical)
    classes (optional) : list
        The unique output classes in order
    normalize (optional) : None, 'pred' or 'class'
        According to which of pred or class we normalize

    Returns
    -------
    numpy ndarray
        Confusion Matrix : shape (num_classes, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import confusion_matrix
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> confusion_matrix(y_pred, y_true, classes=[2, 1, 0])
        2    1    0
    2  1.0  1.0  1.0
    1  0.0  3.0  0.0
    0  1.0  1.0  2.0
    """
    # check and coerce inputs
    _, y_pred, y_true, _, classes = _multiclass_checks(
        p_attr=None,
        y_pred=y_pred,
        y_true=y_true,
        groups=None,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    class_dict = dict(zip(classes, range(num_classes)))

    # initialize the confusion matrix
    confmat = np.zeros((num_classes, num_classes))

    # loop over instances
    for x, y in zip(y_pred, y_true):
        # increment correct entry
        confmat[class_dict[x], class_dict[y]] += 1

    if normalize is None:
        pass

    elif normalize == "pred":
        confmat = confmat / np.sum(confmat, axis=1).reshape(-1, 1)

    elif normalize == "true":
        confmat = confmat / np.sum(confmat, axis=0).reshape(1, -1)

    else:
        raise ValueError('normalize should be one of None, "pred" or "true"')

    return pd.DataFrame(confmat, columns=classes).set_index(np.array(classes))


def confusion_tensor(
    p_attr, y_pred, y_true, groups=None, classes=None, as_tensor=False
):
    """
    Confusion Tensor.

    Description
    ----------
    This function computes the confusion tensor. The k,i,jth
    entry is the number of instances of group k with predicted
    class i and true class j.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector (categorical)
    y_pred : array-like
        Prediction vector (categorical)
    y_true : array-like
        Target vector (categorical)
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order
    as_tensor (optional) : bool, default False
        Whether we return a tensor or DataFrame

    Returns
    -------
    numpy ndarray
        Confusion Tensor : shape (num_groups, num_classes, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import confusion_tensor
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> confusion_tensor(p_attr, y_pred, y_true, as_tensor=True).shape
    (3, 3, 3)
    >>> confusion_tensor(p_attr, y_pred, y_true, as_tensor=True)
    array([[[2., 0., 0.],
            [0., 1., 0.],
            [0., 1., 0.]],

        [[0., 0., 1.],
            [0., 2., 0.],
            [1., 0., 0.]],

        [[0., 1., 0.],
            [0., 0., 0.],
            [0., 0., 1.]]])
    >>> confusion_tensor(p_attr, y_pred, y_true, as_tensor=False)
        A	        B	        C
        0	1	2	0	1	2	0	1	2
    0	2.0	0.0	0.0	0.0	0.0	1.0	0.0	1.0	0.0
    1	0.0	1.0	0.0	0.0	2.0	0.0	0.0	0.0	0.0
    2	0.0	1.0	0.0	1.0	0.0	0.0	0.0	0.0	1.0
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)
    class_dict = dict(zip(classes, range(num_classes)))
    group_dict = dict(zip(groups, range(num_groups)))

    # initialize the confusion tensor
    conftens = np.zeros((num_groups, num_classes, num_classes))

    # loop over instances
    for x, y, z in zip(p_attr, y_pred, y_true):
        # increment correct entry
        conftens[group_dict[x], class_dict[y], class_dict[z]] += 1

    # return as a tensor
    if as_tensor is True:
        return conftens

    # return as pandas DataFrame
    elif as_tensor is False:
        d = {}
        for i, group in enumerate(groups):
            # confusion matrix of group number i
            d[group] = pd.DataFrame(conftens[i, :, :], columns=classes).set_index(
                np.array(classes)
            )
        # create a multilevel pandas dataframe
        multi_df = pd.concat(d, axis=1)
        return multi_df

    else:
        raise ValueError("as_tensor should be boolean")


def frequency_matrix(p_attr, y_pred, groups=None, classes=None, normalize="group"):
    """
    Frequency Matrix.

    Description
    ----------
    This function computes the frequency matrix. For each
    group, class pair we compute the count of that group
    for admission within that class. We include the option to normalise
    over groups or classes. By default we normalise by 'group'.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order
    normalize (optional): None, 'group' or 'class'
        According to which of group or class we normalize

    Returns
    -------
    pandas DataFrame
        Success Rate Matrix : shape (num_groups, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import frequency_matrix
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> frequency_matrix(p_attr, y_pred, normalize='class')
        0     1     2
    A  0.50  0.25  0.25
    B  0.25  0.50  0.25
    C  0.50  0.00  0.50
    """
    # check and coerce inputs
    p_attr, y_pred, _, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)
    class_dict = dict(zip(classes, range(num_classes)))
    group_dict = dict(zip(groups, range(num_groups)))

    # initialize success rate matrix
    sr_mat = np.zeros((num_groups, num_classes))

    # loop over instances
    for x, y in zip(p_attr, y_pred):
        sr_mat[group_dict[x], class_dict[y]] += 1

    # normalize is None, return counts
    if normalize is None:
        return pd.DataFrame(sr_mat, columns=classes).set_index(np.array(groups))

    # normalise over rows
    elif normalize == "group":
        sr_mat = sr_mat / sr_mat.sum(axis=1).reshape(-1, 1)
        return pd.DataFrame(sr_mat, columns=classes).set_index(np.array(groups))

    # normalise over columns
    elif normalize == "class":
        sr_mat = sr_mat / sr_mat.sum(axis=0).reshape(1, -1)
        return pd.DataFrame(sr_mat, columns=classes).set_index(np.array(groups))

    else:
        raise ValueError("normalize has to be one of None, 'group' or 'class'")


def accuracy_matrix(p_attr, y_pred, y_true, groups=None, classes=None):
    """
    Multiclass Accuracy Matrix.

    Description
    ----------
    Given a protected attribute and multiclass classification task,
    for each group and class this function computes the accuracy of
    predictions on that group for the one vs all classifier of that class.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    pandas DataFrame
        Accuracy Matrix : shape (num_groups, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import accuracy_matrix
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> accuracy_matrix(p_attr, y_pred, y_true)
        0    1    2
    A  1.0  0.5  0.0
    B  0.0  1.0  0.0
    C  0.0  0.0  1.0
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)

    # compute confusion tensor
    conftens = confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)

    # initialize accuracy matrix
    acc_mat = np.zeros((num_groups, num_classes))

    # loop over groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :]
        acc_mat[k, :] = np.diag(confmat_k) / (
            confmat_k.sum(axis=0) + confmat_k.sum(axis=1) - np.diag(confmat_k)
        )

    return pd.DataFrame(acc_mat, columns=classes).set_index(np.array(groups))


def precision_matrix(p_attr, y_pred, y_true, groups=None, classes=None):
    """
    Multiclass Precision Matrix.

    Description
    ----------
    Given a protected attribute and multiclass classification task,
    for each group and class this function computes the precision of
    predictions on that group for the one vs all classifier of that class.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    pandas DataFrame
        Precision Matrix : shape (num_groups, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import precision_matrix
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> precision_matrix(p_attr, y_pred, y_true, groups=None, classes=None)
        0    1    2
    A  1.0  0.5  NaN
    B  0.0  1.0  0.0
    C  NaN  0.0  1.0
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)

    # compute confusion tensor
    conftens = confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)

    # initialize precision matrix
    prec_mat = np.zeros((num_groups, num_classes))

    # loop over groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :]
        prec_mat[k, :] = np.diag(confmat_k) / np.sum(confmat_k, axis=0)

    return pd.DataFrame(prec_mat, columns=classes).set_index(np.array(groups))


def recall_matrix(p_attr, y_pred, y_true, groups=None, classes=None):
    """
    Multiclass Recall Matrix.

    Description
    ----------
    Given a protected attribute and multiclass classification task,
    for each group and class this function computes the recall of
    predictions on that group for the one vs all classifier of that class.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    pandas DataFrame
        Recall Matrix : shape (num_groups, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import recall_matrix
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> recall_matrix(p_attr, y_pred, y_true, groups=None, classes=None)
        0    1    2
    A  1.0  1.0  0.0
    B  0.0  1.0  0.0
    C  0.0  NaN  1.0
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)

    # compute confusion tensor
    conftens = confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)

    # initialize recall matrix
    rec_mat = np.zeros((num_groups, num_classes))

    # loop over groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :]
        rec_mat[k, :] = np.diag(confmat_k) / np.sum(confmat_k, axis=1)

    return pd.DataFrame(rec_mat, columns=classes).set_index(np.array(groups))


def multiclass_equality_of_opp(
    p_attr, y_pred, y_true, groups=None, classes=None, aggregation_fun="mean"
):
    """
    Multiclass Equality of Opportunity.

    Description
    ----------
    This metric is a multiclass generalisation of Equality of
    Opportunity. For each group, compute the matrix of error
    rates (normalised confusion matrix). Compute all distances
    (mean absolute deviation) between such matrices. Then
    aggregate them using the mean, or max strategy.

    Interpretation
    --------------
    The accepted values and bounds for this metric are the same
    as the 1d case. A value of 0 is desired. Values below 0.1
    are considered fair.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    scalar
        Multiclass Equality of Opportunity

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import multiclass_equality_of_opp
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> multiclass_equality_of_opp(p_attr, y_pred, y_true)
    0.7962962962962963
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)

    # compute confusion tensor
    conftens = (
        confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)
        + 1e-32
    )
    # normalization constant
    norm = conftens.sum(axis=1)

    # initialize distance matrix
    dist_mat = np.zeros((num_groups, num_groups))

    # distance confusion matrix across groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :] / norm[k, :]
        for j in range(k + 1, num_groups):
            confmat_j = conftens[j, :, :] / norm[j, :]
            dist_mat[k, j] = np.sum(np.abs(confmat_k - confmat_j)) / (2 * num_classes)

    if aggregation_fun == "max":
        res = np.max(dist_mat)
    else:
        res = np.sum(dist_mat) / (num_groups * (num_groups - 1) / 2)

    return res


def multiclass_average_odds(
    p_attr, y_pred, y_true, groups=None, classes=None, aggregation_fun="mean"
):
    """
    Multiclass Average Odds.

    Description
    ----------
    This metric is a multiclass generalisation of Average
    Odds. For each group, compute the matrix of error
    rates (normalised confusion matrix). Average these
    matrices over rows, and compute all pariwise distances
    (mean absolute deviation) between the resulting vectors.
    Aggregate results using either mean or max strategy.

    Interpretation
    --------------
    The accepted values and bounds for this metric are the same
    as the 1d case. A value of 0 is desired. Values below 0.1
    are considered fair.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    scalar
        Multiclass Average Odds

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import multiclass_average_odds
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> multiclass_average_odds(p_attr, y_pred, y_true)
    0.16666666666666666
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    num_groups = len(groups)

    # compute confusion tensor
    conftens = (
        confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)
        + 1e-32
    )
    # normalization constant
    norm = conftens.sum(axis=1)

    # initialize distance matrix
    dist_mat = np.zeros((num_groups, num_groups))

    # distance confusion matrix across groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :] / norm[k, :]
        for j in range(k + 1, num_groups):
            confmat_j = conftens[j, :, :] / norm[j, :]
            dist_mat[k, j] = np.abs((confmat_k - confmat_j).sum(axis=1)).sum() / (
                2 * num_classes
            )

    if aggregation_fun == "max":
        res = np.max(dist_mat)
    else:
        res = np.sum(dist_mat) / (num_groups * (num_groups - 1) / 2)

    return res


def multiclass_true_rates(
    p_attr, y_pred, y_true, groups=None, classes=None, aggregation_fun="mean"
):
    """
    Multiclass True Rates.

    Description
    ----------
    This metric is a multiclass generalisation of TPR
    Difference. For each group, compute the matrix of error
    rates (normalised confusion matrix). Compute all distances
    (mean absolute deviation) between the diagonals of such
    matrices. Then aggregate them using the mean, or max strategy.

    Interpretation
    --------------
    The accepted values and bounds for this metric are the same
    as the 1d case. A value of 0 is desired. Values below 0.1
    are considered fair.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    y_true : array-like
        Target vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order

    Returns
    -------
    pandas DataFrame
        Accuracy Matrix : shape (num_groups, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import multiclass_true_rates
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> multiclass_true_rates(p_attr, y_pred, y_true)
    0.6666666666666666
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=groups,
        classes=classes,
    )

    # variables
    num_groups = len(groups)

    # compute confusion tensor
    conftens = (
        confusion_tensor(p_attr, y_pred, y_true, groups, classes, as_tensor=True)
        + 1e-32
    )
    # normalization constant
    norm = conftens.sum(axis=1)

    # initialize distance matrix
    dist_mat = np.zeros((num_groups, num_groups))

    # distance confusion matrix across groups
    for k in range(num_groups):
        confmat_k = conftens[k, :, :] / norm[k, :]
        for j in range(k + 1, num_groups):
            confmat_j = conftens[j, :, :] / norm[j, :]
            dist_mat[k, j] = np.abs(np.diag(confmat_k - confmat_j)).mean()

    if aggregation_fun == "max":
        res = np.max(dist_mat)
    else:
        res = np.sum(dist_mat) / (num_groups * (num_groups - 1) / 2)

    return res


def multiclass_statistical_parity(
    p_attr, y_pred, groups=None, classes=None, aggregation_fun="mean"
):
    """
    Multiclass statistical parity.

    Description
    ----------
    This function computes statistical parity for a classification task
    with multiple classes and a protected attribute with multiple groups.
    For each group compute the vector of success rates for entering
    each class. Compute all distances (mean absolute deviation) between
    such vectors. Then aggregate them using the mean, or max strategy.

    Interpretation
    --------------
    The accepted values and bounds for this metric are the same
    as the 1d case. A value of 0 is desired. Values below 0.1
    are considered fair.

    Parameters
    ----------
    p_attr : array-like
        Multiclass protected attribute vector.
    y_pred : array-like
        Prediction vector (categorical).
    groups (optional) : list
        Unique groups from p_attr in order
    classes (optional) : list
        The unique output classes in order
    aggregation_fun (optional) : str
        The function to aggregate across groups ('mean' or 'max')

    Returns
    -------
    float
        Multiclass Statistical Parity

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import multiclass_statistical_parity
    >>> p_attr = np.array(['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C'])
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> multiclass_statistical_parity(p_attr, y_pred, aggregation_fun='max')
    0.5
    """
    # check and coerce inputs
    p_attr, y_pred, _, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=groups,
        classes=classes,
    )

    # variables
    num_groups = len(groups)

    # compute frequency matrix (normalised by class)
    sr_mat = frequency_matrix(p_attr, y_pred, groups, classes).to_numpy() + 1e-32

    # initialize distance matrix
    dist_mat = np.zeros((num_groups, num_groups))

    # distance confusion matrix across groups
    for k in range(num_groups):
        pred_prob_k = sr_mat[k]
        for j in range(k + 1, num_groups):
            pred_prob_j = sr_mat[j]
            dist_mat[k, j] = np.abs(pred_prob_k - pred_prob_j).sum() / 2

    if aggregation_fun == "max":
        res = np.max(dist_mat)
    else:
        res = np.sum(dist_mat) / (num_groups * (num_groups - 1) / 2)

    return res


def multiclass_bias_metrics(
    p_attr, y_pred, y_true, groups=None, classes=None, metric_type="equal_outcome"
):
    """
    Multiclass bias metrics batch computation.

    Description
    ----------
    This function computes all the relevant multiclass bias metrics,
    and displays them as a pandas dataframe.

    Parameters
    ----------
    g_min : numpy array
        Minority class vector
    g_maj : numpy array
        Majority class vector
    y_pred : array-like
        Regression predictions vector
    y_true (optional) : numpy array
        Regression target vector
    metric_type : 'both', 'equal_outcome' or 'equal_opportunity'
        Specifies which metrics we compute

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """

    perform = {
        "Max Multiclass Statistical Parity": multiclass_statistical_parity,
        "Mean Multiclass Statistical Parity": multiclass_statistical_parity,
    }

    true_perform = {
        "Max Multiclass Equality of Opportunity": multiclass_equality_of_opp,
        "Max Multiclass Average Odds": multiclass_average_odds,
        "Max Multiclass True Positive Difference": multiclass_true_rates,
        "Mean Multiclass Equality of Opportunity": multiclass_equality_of_opp,
        "Mean Multiclass Average Odds": multiclass_average_odds,
        "Mean Multiclass True Positive Difference": multiclass_true_rates,
    }

    ref_vals = {
        "Max Multiclass Statistical Parity": 0,
        "Mean Multiclass Statistical Parity": 0,
        "Max Multiclass Equality of Opportunity": 0,
        "Max Multiclass Average Odds": 0,
        "Max Multiclass True Positive Difference": 0,
        "Mean Multiclass Equality of Opportunity": 0,
        "Mean Multiclass Average Odds": 0,
        "Mean Multiclass True Positive Difference": 0,
    }

    param = {
        "Max Multiclass Statistical Parity": "max",
        "Mean Multiclass Statistical Parity": "mean",
        "Max Multiclass Equality of Opportunity": "max",
        "Max Multiclass Average Odds": "max",
        "Max Multiclass True Positive Difference": "max",
        "Mean Multiclass Equality of Opportunity": "mean",
        "Mean Multiclass Average Odds": "mean",
        "Mean Multiclass True Positive Difference": "mean",
        "Max Multiclass Statistical Paraity": "max",
        "Mean Multiclass Statistical Paraity": "mean",
    }

    out_metrics = []
    opp_metrics = []

    out_metrics += [
        [
            pf,
            fn(p_attr, y_pred, groups, classes, aggregation_fun=param[pf]),
            ref_vals[pf],
        ]
        for pf, fn in perform.items()
    ]
    if y_true is not None:
        opp_metrics += [
            [
                pf,
                fn(p_attr, y_pred, y_true, groups, classes, aggregation_fun=param[pf]),
                ref_vals[pf],
            ]
            for pf, fn in true_perform.items()
        ]

    if metric_type == "both":
        metrics = out_metrics + opp_metrics
        return pd.DataFrame(
            metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_outcome":
        return pd.DataFrame(
            out_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_opportunity":
        return pd.DataFrame(
            opp_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    else:
        raise ValueError(
            "metric_type is not one of : both, equal_outcome, equal_opportunity"
        )



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator

from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm import FairClusteringAlgorithm


class VariationalFairClustering(BaseEstimator, BMImp):
    """
    Variational Fair Clustering helps you to find clusters with specified proportions
    of different demographic groups pertaining to a sensitive attribute of the dataset
    (group_a and group_b) for any well-known clustering method such as K-means, K-median
    or Spectral clustering (Normalized cut).


    References
    ----------
        Ziko, Imtiaz Masud, et al. "Variational fair clustering." Proceedings of the AAAI
        Conference on Artificial Intelligence. Vol. 35. No. 12. 2021.
    """

    def __init__(
        self,
        n_clusters: Optional[int],
        lipchitz_value: Optional[str] = 1,
        lmbda: Optional[float] = 0.7,
        method: Optional[str] = "kmeans",
        normalize_input: Optional[bool] = True,
        seed: Optional[int] = None,
        verbose: Optional[int] = 0,
    ):
        """
        Parameters
        ----------
            n_clusters : int
                The number of clusters to form as well as the number of centroids to generate.

            lipchitz_value : float
                Lipchitz value in bound update

            lmbda : float
                specified lambda parameter

            method : str
                cluster option : {'kmeans', 'kmedian'} (TODO: 'ncut' take too much time consuming)

            normalize_input : str
                Normalize input data X

            seed : int
                Random seed.

            verbose : bool
                If true , print metrics
        """
        # Constant parameters
        self.algorithm = FairClusteringAlgorithm(
            K=n_clusters,
            L=lipchitz_value,
            lmbda=lmbda,
            method=method,
            normalize_input=normalize_input,
            verbose=verbose,
        )
        self.seed = seed
        self.n_clusters = n_clusters
        self.lipchitz_value = lipchitz_value
        self.lmbda = lmbda
        self.method = method
        self.normalize_input = normalize_input
        self.verbose = verbose
        self.sens_group = SensitiveGroups()

    def fit(
        self,
        X: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair cluster.

        Parameters
        ----------

        X : numpy array
            input matrix

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
        the same object
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]
        p_attr = self.sens_group.fit_transform(
            np.c_[group_a, group_b], convert_numeric=True
        )
        self.algorithm.fit(
            X=X, p_attr=p_attr, random_state=np.random.RandomState(self.seed)
        )
        return self

    @property
    def cluster_centers_(self):
        return self.algorithm.fair_clustering.C

    @property
    def labels_(self):
        return self.algorithm.fair_clustering.l

    def predict(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict cluster for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector


        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]
        p_attr = self.sens_group.transform(
            np.c_[group_a, group_b], convert_numeric=True
        )
        return self.algorithm.predict(X, p_attr)

    def fit_predict(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Prediction

        Description
        ----------
        Fit and Predict the cluster for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector


        Returns
        -------

        numpy.ndarray: Predicted cluster per sample.
        """
        self.fit(X, group_a, group_b)
        return self.labels_



import math

import numpy as np
import pandas as pd


def filter_invalid_examples(rankings, query_col, group_col):
    new_rankings = []
    for q, ranking in rankings.groupby(query_col):
        if (ranking[group_col].sum() > 0).any():
            new_rankings.append(ranking)
    new_rankings = pd.concat(new_rankings, axis=0).reset_index(drop=True)
    return new_rankings


def exposure_metric(rankings, group_col: str, query_col: str, score_col: str):

    rankings = filter_invalid_examples(
        rankings, query_col=query_col, group_col=group_col
    )

    exp_diff = np.mean(
        [
            exposure_diff(ranking[score_col], ranking[group_col])
            for q, ranking in rankings.groupby(query_col)
        ]
    )

    metric_values = [
        exposure_ratio(ranking[score_col], ranking[group_col])
        for q, ranking in rankings.groupby(query_col)
    ]
    exp_ratio = np.mean([m for m in metric_values if not math.isnan(m)])

    df = pd.DataFrame(
        [{"exposure_ratio": exp_ratio, "exposure difference": exp_diff}]
    ).T
    df.columns = ["Value"]
    return df


def exposure_diff(data_per_query, prot_idx_per_query):
    """
    Description
    -----------
    Computes the exposure difference between protected and non-protected groups.

    Parameters
    ----------
    data: matrix-like
        all predictions.

    prot_idx: matrix-like
        list states which item is protected or non-protected.

    Return
    ------
        float value
    """
    (
        judgments_per_query,
        protected_items_per_query,
        nonprotected_items_per_query,
    ) = find_items_per_group_per_query(data_per_query, prot_idx_per_query)

    exposure_prot = normalized_exposure(protected_items_per_query, judgments_per_query)
    exposure_nprot = normalized_exposure(
        nonprotected_items_per_query, judgments_per_query
    )
    exposure_diff = np.maximum(0, (exposure_nprot - exposure_prot))

    return exposure_diff


def exposure_ratio(data_per_query, prot_idx_per_query):
    """
    Description
    -----------
    Computes the exposure difference between protected and non-protected groups.

    Parameters
    ----------
    data: matrix-like
        all predictions.

    prot_idx: matrix-like
        list states which item is protected or non-protected.

    Return
    ------
        float value
    """
    (
        judgments_per_query,
        protected_items_per_query,
        nonprotected_items_per_query,
    ) = find_items_per_group_per_query(data_per_query, prot_idx_per_query)

    exposure_prot = normalized_exposure(protected_items_per_query, judgments_per_query)
    exposure_nprot = normalized_exposure(
        nonprotected_items_per_query, judgments_per_query
    )
    exposure_diff = exposure_nprot / exposure_prot

    return exposure_diff


def find_items_per_group_per_query(data, protected_feature):
    data_per_query = np.array(data).astype(np.float32)
    if np.any(np.isnan(data_per_query)):
        raise Exception(
            "data has NaN values!!, fix your data or normalize it before training."
        )
    if len(data_per_query.shape) == 1:
        data_per_query = data_per_query[:, None]
    protected_feature = np.array(protected_feature)
    protected_items_per_query = data_per_query[protected_feature == True]
    nonprotected_items_per_query = data_per_query[protected_feature == False]
    return data_per_query, protected_items_per_query, nonprotected_items_per_query


def normalized_exposure(group_data, all_data):
    """
    Description
    -----------
    Calculates the exposure of a group in the entire ranking.

    Parameters
    ----------
    group_data: matrix-like
        predictions of relevance scores for one group.

    all_data: matrix-like
        all predictions.

    Return
    ------
    float value that is normalized exposure in a ranking for one group.
    """
    return (np.sum(topp_prot(group_data, all_data) / np.log(2))) / group_data.size


def normalized_topp_prot_deriv_per_group(
    group_features, all_features, group_predictions, all_predictions
):
    """
    Description
    -----------
    Normalizes the results of the derivative of topp_prot.

    Parameters
    ----------
    group_features: array-like
        feature vector of (non-) protected group.

    all_features: array-like
        feature vectors of all data points.

    group_predictions: array-like
        predictions of all data points.

    all_predictions: array-like
        predictions of all data points.

    Return
    numpy array of float values.
    """
    derivative = topp_prot_first_derivative(
        group_features, all_features, group_predictions, all_predictions
    )
    result = (np.sum(derivative / np.log(2), axis=0)) / group_predictions.size
    return result


def topp_prot(group_items, all_items):
    """
    Description
    -----------
    Given a dataset of features what is the probability of being at the top position.

    Parameters
    ----------
    group_items:
        Vector of predicted scores of one group (protected or non-protected).

    all_items:
        Vector of predicted scores of all items.

    Return
    ------
        Numpy array of float values.
    """
    return np.exp(group_items) / np.sum(np.exp(all_items))


def topp_prot_first_derivative(
    group_features, all_features, group_predictions, all_predictions
):
    """
    Description
    -----------
    Derivative for topp_prot in pieces.

    Parameters
    ----------
    group_features: array-like
        Feature vector of (non-) protected group.

    group_predictions: array-like
        Predicted scores for (non-) protected group.

    all_predictions: array-like
        Predictions of all data points.

    all_features: array-like
        Feature vectors of all data points.

    Return
    ------
    numpy array with weight adjustments.
    """

    numerator1 = np.dot(np.transpose(np.exp(group_predictions)), group_features)
    numerator2 = np.sum(np.exp(all_predictions))
    numerator3 = np.sum(np.dot(np.transpose(np.exp(all_predictions)), all_features))
    denominator = np.sum(np.exp(all_predictions)) ** 2

    result = (
        numerator1 * numerator2 - np.exp(group_predictions) * numerator3
    ) / denominator

    # return result as flat numpy array instead of matrix
    return np.asarray(result)


def topp(v):
    """
    Description
    -----------
    Computes the probability of a document being.

    Parameters
    ----------
    v: array-like
        all training judgments or all predictions.

    Return:
    float value which is a probability.

    """
    return np.exp(v) / np.sum(np.exp(v))


def hh(q, x):
    return (q, hash(str(x)))


class Standarizer:
    def __init__(self, group_col):
        self.group_col = group_col

    def fit_transform(self, feature_matrix):
        self._mus = np.array(feature_matrix).mean()
        self._sigmas = np.array(feature_matrix).std()
        protected_feature = feature_matrix[self.group_col]
        feature_matrix = (feature_matrix - self._mus) / self._sigmas
        feature_matrix[self.group_col] = protected_feature
        return feature_matrix

    def transform(self, feature_matrix):
        protected_feature = feature_matrix[self.group_col]
        feature_matrix = (feature_matrix - self._mus) / self._sigmas
        feature_matrix[self.group_col] = protected_feature
        return feature_matrix



import copy
import sys
from typing import Optional

import numpy as np

from ._grid_generator import GridGenerator


class GridSearchAlgorithm:
    """
    Estimator to perform a grid search given a blackbox estimator algorithm.
    The approach used is taken from section 3.4 of "A reductions approach to fair classification"
    Enabled for more than two sensitive feature values

    Reference
    ---------

    Agarwal, Alekh, et al. "A reductions approach to fair classification."
    International Conference on Machine Learning. PMLR, 2018.
    """

    def __init__(
        self,
        estimator,
        constraint,
        constraint_weight: Optional[float] = 0.5,
        grid_size: Optional[int] = 20,
        grid_limit: Optional[int] = 2,
        verbose: Optional[int] = 0,
    ):
        """
        Init GridSearchAlgorithm object

        Parameters
        ----------

        estimator :
            An estimator implementing methods :code:`fit(X, y, sample_weight=sample_weight)` and
            :code:`predict(X)`, where `X` is the matrix of features, `y` is the
            vector of labels, and `sample_weight` is a vector of weights.
            In binary classification labels `y` and predictions returned by
            :code:`predict(X)` are either 0 or 1.

        generator : float
            grid generator utility

        constraint : ClassificationContraint
            The disparity constraint

        constraint_weight : float
            Specifies the relative weight put on the constraint violation when selecting the
            best model. The weight placed on the error rate will be :code:`1-constraint_weight`

        grid_size: int
            number of columns to be generated in the grid.

        grid_limit : float
            range of the values in the grid generated.

        verbose : int
            If >0, will show progress percentage.
        """

        self.constraint = constraint
        self.estimator = estimator
        self.objective = constraint.default_objective()
        self.grid_limit = grid_limit
        self.grid_size = grid_size
        self.monitor = Monitor(constraint_weight=constraint_weight, verbose=verbose)

    def fit(self, X, y, sensitive_features):
        """
        Fit model using Grid Search Algorithm.

        Parameters
        ----------

        X : matrix-like
            input matrix

        y_true : numpy array
            target vector

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]

        Returns
        -------
        the same object
        """
        self.constraint.load_data(X, y, sensitive_features)
        self.objective.load_data(X, y, sensitive_features)

        neg_allowed = self.constraint.neg_basis_present
        objective_in_the_span = self.constraint.default_objective_lambda_vec is not None

        self.generator = GridGenerator(
            grid_size=self.grid_size,
            grid_limit=self.grid_limit,
            neg_allowed=neg_allowed,
            force_L1_norm=objective_in_the_span,
        )

        grid = self.generator.generate_grid(self.constraint)
        self.monitor.total_steps = grid.shape[1]
        for (_, lambda_vec) in grid.iteritems():

            weights = self.constraint.signed_weights(lambda_vec)
            if not objective_in_the_span:
                weights += self.objective.signed_weights()

            if self.constraint.PROBLEM_TYPE == "classification":
                y_reduction = 1 * (weights > 0)
                weights = weights.abs()
            else:
                y_reduction = self.constraint._y_as_series

            current_estimator = copy.deepcopy(self.estimator)
            current_estimator.fit(X, y_reduction, sample_weight=weights)

            predict_fn = lambda X: current_estimator.predict(X)
            objective_ = self.objective.gamma(predict_fn)[0]
            gamma_ = self.constraint.gamma(predict_fn)

            self.monitor.save(lambda_vec, current_estimator, objective_, gamma_)
            self.monitor.log_progress()

        self.best_idx_ = self.monitor.get_best_idx()

        return self

    def predict(self, X):
        return self.monitor.predictors_[self.best_idx_].predict(X)

    def predict_proba(self, X):
        return self.monitor.predictors_[self.best_idx_].predict_proba(X)


class Monitor:
    """Monitor class used to store historical data"""

    def __init__(self, constraint_weight, verbose):
        self.predictors_ = []
        self.lambda_vecs_ = []
        self.objectives_ = []
        self.gammas_ = []
        self.losses = []
        self.constraint_weight = float(constraint_weight)
        self.objective_weight = 1.0 - constraint_weight
        self.verbose = verbose
        self.total_steps = 0
        self.step = 0

    def loss_fct(self, objective, gamma):
        return self.objective_weight * objective + self.constraint_weight * gamma.max()

    def save(self, lambda_vec, current_estimator, objective, gamma):

        self.predictors_.append(current_estimator)
        self.losses.append(self.loss_fct(objective, gamma))
        self.objectives_.append(objective)
        self.gammas_.append(gamma)
        self.lambda_vecs_.append(lambda_vec)

    def get_best_idx(self):
        return self.losses.index(min(self.losses))

    def log_progress(self):
        self.step += 1
        if self.verbose:
            sys.stdout.write(
                f"\r{self.step}/{self.total_steps}\tloss (best):{min(self.losses):.4f}"
            )
            sys.stdout.flush()



import torch
from torch import nn
from torch.autograd import Variable


class ClassifierModel(nn.Module):
    "Classifier Model, You can change the layer configuration as you wish!"

    def __init__(self, features_dim, hidden_size, keep_prob):
        super().__init__()
        self.hidden1 = nn.Linear(features_dim, hidden_size)
        self.hidden2 = nn.Linear(hidden_size, 1)
        self.act = nn.ReLU()
        self.drop = nn.Dropout(keep_prob)
        self.out_act = nn.Sigmoid()

    def forward(self, x, trainable=True):
        x = self.hidden1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.hidden2(x)
        p = self.out_act(x)
        if trainable:
            return p, x
        else:
            return p


class AdversarialModel(nn.Module):
    """
    Adversarial Model, proposed by B. H. Zhang et al.

    Reference:
        B. H. Zhang, B. Lemoine, and M. Mitchell, "Mitigating Unwanted
        Biases with Adversarial Learning," AAAI/ACM Conference on Artificial
        Intelligence, Ethics, and Society, 2018.
    """

    def __init__(self):
        super().__init__()
        self.c = Variable(
            torch.randn(
                1,
            ).type(torch.FloatTensor),
            requires_grad=True,
        )
        self.act = nn.Sigmoid()
        self.hidden = nn.Linear(3, 1)
        self.out_act = nn.Sigmoid()

    def forward(self, x, y, trainable=True):
        s = self.act((1 + torch.absolute(self.c)) * x)
        x = torch.cat((s, s * y, s * (1.0 - y)), 1)
        x = self.hidden(x)
        p = self.out_act(x)
        if trainable:
            return p, x
        else:
            return p


class ADModel(nn.Module):
    """
    Complete system integrate classifier and adversarial submodels.
    """

    def __init__(self, estimator: nn.Module):
        super().__init__()
        self.classifier = estimator
        self.adversarial = AdversarialModel()

    def forward(self, x, y):
        y_prob, y_logits = self.classifier(x)
        z_prob, _ = self.adversarial(y_logits, y)
        return y_prob, z_prob



import random
from collections import defaultdict

import networkx as nx


class CategoricalFeature:
    def __init__(self, data, name="no_name"):
        self.data = data
        self.name = name
        d1 = defaultdict(int)  # bin_data
        d2 = defaultdict(int)  # bin_index_dict
        d3 = defaultdict(list)  # bin_fulldata
        d4 = defaultdict(int)  # bin_index_dict_reverse
        d5 = defaultdict(int)  # category_count

        n = len(self.data)
        count = 0
        for i in range(0, n):
            obs = self.data[i]
            if obs in d2:
                pass  # if obs (i.e. category) is alreay a KEY in bin_index_data then don't do anything
            else:
                d2[
                    obs
                ] = count  # bin_index_dict inits the KEY: category, with VALUE: count
                d4[count] = obs  # bin_index_dict_reverse does the opposite
                count += 1
            bin_idx = d2[obs]
            d1[bin_idx] += 1  # add 1 to the obs category idex in bin_data
            d5[obs] += 1  # add 1 to the obs category NAME in category_count
            d3[bin_idx].append(
                i
            )  # add obs to the list of obs with that category in bin_fulldata

        self.bin_data = d1
        self.category_count = d5
        self.num_bins = len(list(d1.items()))
        self.bin_fulldata = d3
        self.bin_index_dict = d2
        self.bin_index_dict_reverse = d4

    def create_graph(self, count_generator):
        """
        Description
        -----------
        Creates graph given a CategoricalFeature object.
        """
        DG = nx.DiGraph()  # using networkx package
        bin_list = list(self.bin_data.items())
        bin_index_dict_reverse = self.bin_index_dict_reverse
        k = self.num_bins
        total_items = 0
        for i, amt in bin_list:
            total_items += amt
        DG.add_node("s")
        DG.add_node("t")
        for i in range(
            0, k
        ):  # lefthand side nodes have capacity = number of observations in category i
            DG.add_node(i)
            DG.add_edge("s", i, capacity=bin_list[i][1], weight=0)
        for i in range(
            k, 2 * k
        ):  # righthand side nodes have capacity = DESIRED number of observations in category i
            DG.add_node(i)
            cat = bin_index_dict_reverse[i - k]
            desired_count = count_generator(cat)
            DG.add_edge(i, "t", capacity=desired_count, weight=0)
        # Add special node to hold overflow
        DG.add_node(2 * k)
        DG.add_edge(2 * k, "t", capacity=total_items, weight=0)
        for i in range(0, k):
            for j in range(
                k, 2 * k
            ):  # for each edge from a lefthand side node to a righhand side node:
                if (
                    i + k
                ) == j:  # IF they represent the same category, the edge weight is 0
                    DG.add_edge(i, j, weight=0)
                else:  # IF they represent different categories, the edge weight is 1
                    DG.add_edge(i, j, weight=1)
            DG.add_edge(i, 2 * k, weight=2)
        return DG

    def repair(self, DG):
        mincostFlow = nx.max_flow_min_cost(
            DG, "s", "t"
        )  # max_flow_min_cost returns Dictionary of dictionaries. Keyed by nodes such that mincostFlow[u][v] is the flow edge (u,v)
        bin_dict = self.bin_fulldata
        index_dict = self.bin_index_dict_reverse
        size_data = len(self.data)
        repair_bin_dict = {}
        repair_data = [
            0
        ] * size_data  # initialize repaired data to be 0. If there are zero's after we fill it in the those observations belong in the overflow, "no category"
        k = self.num_bins
        overflow = 0
        for i in range(0, k):  # for each lefthand side node i
            overflow += mincostFlow[i][2 * k]
            for j in range(k, 2 * k):  # for each righthand side node j
                edgeflow = mincostFlow[i][
                    j
                ]  # get the int (edgeflow) representing the amount of observations going from node i to j
                group = random.sample(
                    bin_dict[i], int(edgeflow)
                )  # randomly sample x (edgeflow) unique elements from the list of observations in that category.
                q = j - k  # q is the category index for a righhand side node
                for (
                    elem
                ) in group:  # for each element in the randomly selected group list
                    bin_dict[i].remove(
                        elem
                    )  # remove the element from the list of observation in that category
                    repair_data[elem] = index_dict[
                        q
                    ]  # Mutate repair data at the index of the observation (elem) with its new category (it was 0) which is the category index for the righthand side node it flows to
                if q in repair_bin_dict:  # if the category index is already keyed
                    repair_bin_dict[q].extend(
                        group
                    )  # extend the list of observations with a new list of observations in that category
                else:
                    repair_bin_dict[
                        q
                    ] = group  # otherwise key that category index and set it's value as the group list in that category
        new_feature = CategoricalFeature(
            repair_data
        )  # initialize our new_feature (repaired feature)
        new_feature.bin_fulldata = repair_bin_dict
        return new_feature, overflow



"""
The :mod:`holisticai.utils` module includes utils helper tools
"""

# formatting
from ._formatting import (
    extract_columns,
    extract_group_vectors,
    mat_to_binary,
    normalize_tensor,
    recommender_formatter,
)

# plotting
from ._plotting import get_colors

__all__ = [
    "extract_columns",
    "mat_to_binary",
    "normalize_tensor",
    "get_colors",
    "recommender_formatter",
    "extract_group_vectors",
]



# Base Imports
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

# sklearn imports
from sklearn.metrics import mean_absolute_error, mean_squared_error

# utils
from holisticai.utils import get_colors

from ...utils._validation import _multiclass_checks, _regression_checks


def success_rate_curve(group_a, group_b, y_pred, ax=None, size=None, title=None):
    """
    Success rate A vs B curve.

    Description
    -----------
    This function plots the group_a pass rate
    vs group_b pass rate curve.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # calculation
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Sucess rate A vs B curve")

    # charting
    colors = get_colors(1)
    ax.plot(np.array(pass_b) * 100.0, np.array(pass_a) * 100.0, color=colors[0])
    ax.plot([0, 100], [0, 100], linestyle="--", label="ideal curve", color="grey")
    ax.set_xlabel("Group B Success Rate %")
    ax.set_ylabel("Group A Success Rate %")
    ax.legend()

    return ax


def statistical_parity_curve(
    group_a, group_b, y_pred, x_axis="score", ax=None, size=None, title=None
):
    """
    Statistical Parity Curve

    Description
    -----------
    This function plots the statistical parity versus
    threshold curves for min and maj groups.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    x_axis (optional) : 'score' or 'quantile'
        Indicates whether the x axis is the scores
        or data quantiles
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # calculations
    thresh = np.linspace(0, 1, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()

    # setup
    sns.set()
    colors = get_colors(1)
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Statistical Parity Curve")

    # charting
    if x_axis == "score":
        ax.plot(
            pass_value,
            pass_a - pass_b,
            color=colors[0],
            label="statistical parity curve",
        )
        ax.plot(
            [np.min(pass_value), np.max(pass_value)],
            [0.1, 0.1],
            linestyle="--",
            label="upper bound",
            color="grey",
        )
        ax.plot(
            [np.min(pass_value), np.max(pass_value)],
            [-0.1, -0.1],
            linestyle="--",
            label="lower bound",
            color="grey",
        )
        ax.set_xlabel("Score")
        ax.set_ylabel("Success Rate A - Sucess Rate B")
        ax.legend()
    elif x_axis == "quantile":
        ax.plot(
            thresh, pass_a - pass_b, color=colors[0], label="statistical parity curve"
        )
        ax.plot([0, 1], [0.1, 0.1], linestyle="--", label="upper bound", color="grey")
        ax.plot([0, 1], [-0.1, -0.1], linestyle="--", label="lower bound", color="grey")
        ax.set_xlabel("Quantile")
        ax.set_ylabel("Success Rate A - Sucess Rate B")
        ax.legend()

    else:
        raise ValueError("x_axis is not one of : quantile, score")

    return ax


def disparate_impact_curve(
    group_a, group_b, y_pred, x_axis="score", ax=None, size=None, title=None
):
    """
    Disparate Impact curve

    Description
    -----------
    This function plots the Disparate Impact versus threshold curve.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    x_axis (optional) : 'score' or 'quantile'
        Indicates whether the x axis is the scores
        or data quantiles
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # calculations
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()

    # setup
    sns.set()
    colors = get_colors(1)
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Disparate Impact Curve")

    # charting
    if x_axis == "score":
        ax.plot(
            pass_value, pass_a / pass_b, color=colors[0], label="disparate impact curve"
        )
        ax.plot(
            [np.min(pass_value), np.max(pass_value)],
            [0.8, 0.8],
            linestyle="--",
            color="grey",
            label="upper bound",
        )
        ax.plot(
            [np.min(pass_value), np.max(pass_value)],
            [1.2, 1.2],
            linestyle="--",
            color="grey",
            label="lower bound",
        )
        ax.set_xlabel("Score")
        ax.set_ylabel("Disparate Impact")
        ax.legend()

    elif x_axis == "quantile":
        ax.plot(
            thresh, pass_a / pass_b, color=colors[0], label="disparate impact curve"
        )
        ax.plot([0, 1], [0.8, 0.8], linestyle="--", color="grey", label="upper bound")
        ax.plot([0, 1], [1.2, 1.2], linestyle="--", color="grey", label="lower bound")
        ax.set_xlabel("Quantile")
        ax.set_ylabel("Disparate Impact")
        ax.legend()

    else:
        raise ValueError("x_axis is not one of : score, quantile")

    return ax


def success_rate_curves(
    p_attr, y_pred, groups=None, x_axis="score", ax=None, size=None, title=None
):
    """
    Success Rate Curve

    Description
    -----------
    This function plots the success rate vs threshold curve
    for each group.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Predictions vector (regression)
    groups (optional) : list
        The groups we are considering
    x_axis (optional) : 'score' or 'quantile'
        Indicates whether the x axis is the scores
        or data quantiles
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, _, _, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=None,
        classes=None,
    )

    # calculations
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    if groups is None:
        groups = np.sort(np.unique(p_attr))

    # setup
    colors = get_colors(len(groups) + 1, extended_colors=True)
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Success Rate Curves")

    # charting
    if x_axis == "score":
        for i, c in enumerate(groups):
            members = p_attr == c
            pass_c = y_binary[members].sum(axis=0) / members.sum()
            ax.plot(
                pass_value,
                np.array(pass_c) * 100.0,
                label="SR_" + str(c),
                color=colors[i + 1],
            )
        # tot plot
        pass_tot = y_binary.sum(axis=0) / len(p_attr)
        ax.plot(pass_value, np.array(pass_tot) * 100.0, label="SR_tot", color="grey")
        ax.set_xlabel("Score")
        ax.legend()

    elif x_axis == "quantile":
        for i, c in enumerate(groups):
            members = p_attr == c
            pass_c = y_binary[members].sum(axis=0) / members.sum()
            ax.plot(
                thresh,
                np.array(pass_c) * 100.0,
                label="SR_" + str(c),
                color=colors[i + 1],
            )
        # tot plot
        pass_tot = y_binary.sum(axis=0) / len(p_attr)
        ax.plot(thresh, np.array(pass_tot) * 100.0, label="SR_tot", color="grey")
        ax.set_xlabel("Quantile")
        ax.legend()

    else:
        raise ValueError("x_axis is not one of : score, quantile")

    return ax


def rmse_bar_plot(p_attr, y_pred, y_true, ax=None, size=None, title=None):
    """
    RMSE Bar Plot

    Description
    -----------
    This function plots the RMSE for each group
    as a bar plot.


    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Predictions vector (regression)
    y_true : array-like
        Target vector (regression)
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, _, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=None,
        classes=None,
    )

    # extract groups
    groups = np.sort(np.unique(p_attr))
    rmse_list = []

    # loop over groups
    for c in groups:
        members = p_attr == c
        pred_c = y_pred[members]
        true_c = y_true[members]
        rmse_list.append(mean_squared_error(true_c, pred_c, squared=False))
    # tot
    rmse_tot = mean_squared_error(y_true, y_pred, squared=False)
    rmse_list.append(rmse_tot)
    groups = list(groups)
    groups.append("Total")

    # sort
    rmse_list_sorted, groups_sorted = zip(*sorted(zip(rmse_list, groups)))
    rmse_list_sorted, groups_sorted = list(rmse_list_sorted), list(groups_sorted)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("RMSE bar plot")

    # charting
    colors = get_colors(len(groups))
    hai_palette = sns.color_palette(colors)
    ax.set_xlabel("Group")
    ax.set_ylabel("RMSE")
    sns.barplot(x=groups_sorted, y=rmse_list_sorted, palette=hai_palette)
    _, labels = plt.xticks()
    plt.setp(labels, rotation=45)

    return ax


def mae_bar_plot(p_attr, y_pred, y_true, ax=None, size=None, title=None):
    """
    MAE Bar Plot

    Description
    -----------
    This function plots the MAE for each group
    as a bar plot.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Predictions vector (regression)
    y_true : array-like
        Target vector (regression)
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, _, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=None,
        classes=None,
    )

    # extract groups
    groups = np.sort(np.unique(p_attr))
    mae_list = []

    # loop over groups
    for c in groups:
        members = p_attr == c
        pred_c = y_pred[members]
        true_c = y_true[members]
        mae_list.append(mean_absolute_error(true_c, pred_c))

    # tot
    mae_tot = mean_absolute_error(y_true, y_pred)
    mae_list.append(mae_tot)
    groups = list(groups)
    groups.append("Total")

    # sort
    mae_list_sorted, groups_sorted = zip(*sorted(zip(mae_list, groups)))
    mae_list_sorted, groups_sorted = list(mae_list_sorted), list(groups_sorted)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("MAE bar plot")

    # charting
    colors = get_colors(len(groups))
    hai_palette = sns.color_palette(colors)
    ax.set_xlabel("Group")
    ax.set_ylabel("MAE")
    sns.barplot(x=groups_sorted, y=mae_list_sorted, palette=hai_palette)
    _, labels = plt.xticks()
    plt.setp(labels, rotation=45)

    return ax



import itertools
import sys
from typing import Optional, Tuple, Union

import numpy as np
from sklearn.metrics import balanced_accuracy_score

from holisticai.bias import metrics
from holisticai.utils.transformers.bias import BMPostprocessing as BMPost


class RejectOptionClassification(BMPost):
    """
    Reject option classification gives favorable outcomes (y=1) to unpriviliged groups and unfavorable outcomes (y=0)to
    priviliged groups in a confidence band around the decision boundary with the highest uncertainty.

    References:
    Kamiran, Faisal, Asim Karim, and Xiangliang Zhang. "Decision theory for discrimination-aware classification."
    2012 IEEE 12th International Conference on Data Mining. IEEE, 2012.
    """

    ALLOWED_METRICS = [
        "Statistical parity difference",
        "Average odds difference",
        "Equal opportunity difference",
    ]

    def __init__(
        self,
        low_class_thresh: Optional[float] = 0.01,
        high_class_thresh: Optional[float] = 0.99,
        num_class_thresh: Optional[int] = 100,
        num_ROC_margin: Optional[int] = 50,
        metric_name: Optional[str] = "Statistical parity difference",
        metric_ub: Optional[float] = 0.05,
        metric_lb: Optional[float] = -0.05,
        num_workers: Optional[int] = 8,
        verbose: Optional[int] = 0,
    ):
        """
        Create a Reject Option Classification Post-processing instance.

        Parameters
        ----------
        low_class_thresh : float
            Smallest classification threshold to use in the optimization.
            Should be between 0. and 1.
        low_class_thresh : float
            Smallest classification threshold to use in the optimization.
            Should be between 0. and 1.
        high_class_thresh : float
            Highest classification threshold to use in the optimization.
            Should be between 0. and 1.
        num_class_thresh : int
            Number of classification thresholds between low_class_thresh and high_class_thresh for the optimization
            search. Should be > 0.
        num_ROC_margin : int
            Number of relevant ROC margins to be used in the optimization search. Should be > 0.
        metric_name : str
            Name of the metric to use for the optimization. Allowed options are:
            "Statistical parity difference",
            "Average odds difference",
            "Equal opportunity difference".
        metric_ub : float
            Upper bound of constraint on the metric value
        metric_lb : float
            Lower bound of constraint on the metric value
        verbose : int
            If >0, will show progress percentage.

        """
        super().__init__()

        allowed_metrics = [
            "Statistical parity difference",
            "Average odds difference",
            "Equal opportunity difference",
        ]

        self.low_class_thresh = low_class_thresh
        self.high_class_thresh = high_class_thresh
        self.num_class_thresh = num_class_thresh
        self.num_ROC_margin = num_ROC_margin
        self.metric_name = metric_name
        self.metric_ub = metric_ub
        self.metric_lb = metric_lb
        self.num_workers = num_workers
        self.verbose = verbose

        self.classification_threshold = None
        self.ROC_margin = None

        if (
            (self.low_class_thresh < 0.0)
            or (self.low_class_thresh > 1.0)
            or (self.high_class_thresh < 0.0)
            or (self.high_class_thresh > 1.0)
            or (self.low_class_thresh >= self.high_class_thresh)
            or (self.num_class_thresh < 1)
            or (self.num_ROC_margin < 1)
        ):

            raise ValueError("Input parameter values out of bounds")

        if metric_name not in allowed_metrics:
            raise ValueError("metric name not in the list of allowed metrics")

    def fit(
        self,
        y_true: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Compute parameters for reject option classification strategy.

        Description
        ----------
        Evaluate the likelihood with different thresholds. Select the threshold and ROC margin \
        with the best fair metric value and highest balanced accuracy.

        Parameters
        ----------            
        y_true : array-like
            Target vector (nb_examlpes,)        
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability 
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
            
        Returns
        -------
        Self
        """

        params = self._load_data(
            y_true=y_true, y_proba=y_proba, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_true = params["y_true"]
        likelihoods = params["y_score"]
        y_proba = params["y_proba"]

        class_thresholds = np.linspace(
            self.low_class_thresh, self.high_class_thresh, self.num_class_thresh
        )

        args_iterator = [
            (
                self.metric_name,
                y_true,
                likelihoods,
                group_a,
                group_b,
                class_threshold,
                self.num_ROC_margin,
            )
            for class_threshold in class_thresholds
        ]

        # Serial
        configurations = []
        self.num_cases = len(args_iterator)
        for i, argv in enumerate(args_iterator):
            configurations.append(evaluate_threshold(*argv))
            self._log_progres(i)

        # Pool
        # from multiprocessing import Pool
        # with Pool(self.num_workers) as p:
        #    configurations = list(p.starmap(evaluate_threshold, args_iterator))

        # Joblib
        # from joblib import Parallel, delayed
        # configurations = Parallel(n_jobs=self.num_workers, verbose=0)(
        #    delayed(evaluate_threshold)(*args) for args in args_iterator
        # )

        configurations = list(itertools.chain.from_iterable(configurations))

        selected_configurations = list(
            filter(
                lambda c: (c["fair_metric"] >= self.metric_lb)
                and (c["fair_metric"] <= self.metric_ub),
                configurations,
            )
        )

        if any(selected_configurations):
            best_config = max(
                selected_configurations, key=lambda c: c["balanced_accurracy"]
            )
        else:
            best_config = min(configurations, key=lambda c: abs(c["fair_metric"]))

        self.ROC_margin = best_config["roc_margin"]
        self.classification_threshold = best_config["class_thresh"]

        return self

    def transform(
        self,
        y_pred: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Obtain fair predictions using the ROC method.

        Description
        ----------
        Predict the output using the fitted threshold and ROC margin.

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dictionnary with new predictions
        """
        params = self._load_data(
            y_pred=y_pred, y_proba=y_proba, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        y_score = params["y_score"]

        new_y_pred = predict(
            y_score,
            group_a,
            group_b,
            self.classification_threshold,
            self.ROC_margin,
        )
        new_y_score = y_score.copy()

        return {
            "y_pred": new_y_pred,
            "y_score": new_y_score,
        }

    def _log_progres(self, i):
        if self.verbose > 0:
            sys.stdout.write(f"\rProgress: {i+1}/{self.num_cases}")
            sys.stdout.flush()


def evaluate_threshold(
    metric_name, labels, likelihoods, group_a, group_b, class_thresh, num_roc_margin
):
    configurations = []
    high_roc_margin = class_thresh if class_thresh <= 0.5 else 1.0 - class_thresh
    for roc_margin in np.linspace(0.0, high_roc_margin, num_roc_margin):
        # Predict using the current threshold and margin
        predictions = predict(likelihoods, group_a, group_b, class_thresh, roc_margin)

        if metric_name == "Statistical parity difference":
            fair_metric = metrics.statistical_parity(group_b, group_a, predictions)

        elif metric_name == "Average odds difference":
            fair_metric = metrics.average_odds_diff(
                group_b, group_a, predictions, labels
            )

        elif metric_name == "Equal opportunity difference":
            fair_metric = metrics.equal_opportunity_diff(
                group_b, group_a, predictions, labels
            )

        configurations.append(
            {
                "class_thresh": class_thresh,
                "roc_margin": roc_margin,
                "balanced_accurracy": balanced_accuracy_score(labels, predictions),
                "fair_metric": fair_metric,
            }
        )
    return configurations


def predict(likelihoods, group_a, group_b, threshold, roc_margin):
    # Indices of critical region around the classification boundary
    upper_threshold = threshold + roc_margin
    lower_threshold = threshold - roc_margin
    crit_region_inds = np.logical_and(
        likelihoods <= upper_threshold, likelihoods > lower_threshold
    )

    # New, fairer labels
    new_predictions = np.where(likelihoods > threshold, 1, 0)
    new_predictions[np.logical_and(crit_region_inds, group_a)] = 1
    new_predictions[np.logical_and(crit_region_inds, group_b)] = 0
    return new_predictions



from abc import abstractmethod
from itertools import combinations

import numpy as np


class Losses:
    def __call__(self, dr, overall_rates):
        # Getting the overall error rates and group proportions
        s = overall_rates.acc
        e = 1 - s

        # Setting up the coefficients for the objective function
        obj_coefs = np.array([[(s - e) * r[0], (e - s) * r[1]] for r in dr]).flatten()
        return obj_coefs


class ConstraintBase:
    @abstractmethod
    def get_conditions(self):
        pass

    def _get_rates(self, group_rates, groups):
        # Generating the pairs for comparison
        n_groups = len(groups)
        group_combos = list(combinations(groups, 2))
        id_combos = list(combinations(range(n_groups), 2))

        # Pair drop to keep things full-rank with 3 or more groups
        if n_groups > 2:
            n_comp = n_groups - 1
            group_combos = group_combos[:n_comp]
            id_combos = id_combos[:n_comp]

        col_combos = np.array(id_combos) * 2
        n_pairs = len(group_combos)

        # Making empty matrices to hold the pairwise constraint coefficients
        tprs = np.zeros(shape=(n_pairs, 2 * n_groups))
        fprs = np.zeros(shape=(n_pairs, 2 * n_groups))

        # Filling in the constraint matrices
        for i, cols in enumerate(col_combos):
            # Fetching the group-specific rates
            gc = group_combos[i]
            g0 = group_rates[gc[0]]
            g1 = group_rates[gc[1]]

            # Filling in the group-specific coefficients
            tprs[i, cols[0]] = g0.fnr
            tprs[i, cols[0] + 1] = g0.tpr
            tprs[i, cols[1]] = -g1.fnr
            tprs[i, cols[1] + 1] = -g1.tpr

            fprs[i, cols[0]] = g0.tnr
            fprs[i, cols[0] + 1] = g0.fpr
            fprs[i, cols[1]] = -g1.tnr
            fprs[i, cols[1] + 1] = -g1.fpr

        # Choosing whether to go with equalized odds or opportunity
        return tprs, fprs


class EqualizedOdds(ConstraintBase):
    def get_conditions(self, group_rates, groups):
        tprs, fprs = self._get_rates(group_rates, groups)
        con = np.vstack((tprs, fprs))
        con_bounds = np.zeros(con.shape[0])
        return con, con_bounds


class EqualizedOpportunity(ConstraintBase):
    def get_conditions(self, group_rates, groups):
        tprs, fprs = self._get_rates(group_rates, groups)
        con = tprs
        con_bounds = np.zeros(con.shape[0])
        return con, con_bounds



import numpy as np

from .algorithm_utils import FullGDDebiaser, RTLogger, SGDDebiaser


class RandomizedThresholdAlgorithm:
    """
    Threshold optimizer (RTO) to debias models via postprocessing.
    Debias predictions w.r.t. the sensitive class in each demographic group.
    This procedure takes as input a vector y and solves the optimization
    problem subject to the statistical parity constraint.

    Reference
    ---------
    Alabdulmohsin, Ibrahim M., and Mario Lucic. "A near-optimal algorithm for debiasing
    trained machine learning models." Advances in Neural Information Processing Systems
    34 (2021): 8072-8084.
    """

    def __init__(
        self,
        gamma=1.0,
        eps=0.0,
        rho=None,
        sgd_steps=10_000,
        full_gradient_epochs=1_000,
        batch_size=256,
        verbose=False,
    ):
        """
        Instantiate object.

        Parameters
        ----------
          gamma: float
            The regularization parameter gamma (for randomization). Set this to
            1 if the goal is to minmize changes to the original scores.

          eps: float
            Tolerance parameter for bias between 0 and 1 inclusive.

          rho: float
            The rho parameter in the post-hoc rule. If None, rho = E[y].

          sgd_steps: int
            Number of minibatch steps in SGD.

          full_gradient_epochs: int
            Number of epochs in full gradient descent phase.

          batch_size: int
            Size of minibatches in SGD.

          verbose: bool
            If True, display progress.
        """
        self.num_groups = 1
        self.gamma = gamma
        self.eps = eps
        self.rho = rho
        self.sgd_steps = sgd_steps
        self.full_gradient_epochs = full_gradient_epochs
        self.batch_size = batch_size
        self.verbose = verbose

    def fit(self, y_score, groups_num):
        """
        Run the debiaser optimzier

        Description
        -----------
          minimize_x   gamma/2 ||x||^2 - y^Tx
          s.t.      x satisfies DP constraints with tolerance eps and parameter rho.
        The optimizer proceeds in two rounds:
          - First is SGD.
          - Second is full gradient descent.

        Parameters
        ----------
          y_score: np.ndarray
            Predicted probability vector

          groups_num: np.ndarray
            Group membership vector from group 0 to group (num_classes-1).
        Returns:
          None.
        """
        self.yscale = "positive" if min(y_score) >= 0 else "negative"
        self.avrg_y_score = float(sum(y_score)) / len(y_score)
        if self.rho is None:
            if self.yscale == "positive":
                self.rho = self.avrg_y_score
            else:
                self.rho = self.avrg_y_score / 2.0 + 0.5

        if self.rho <= 0:
            raise ValueError("rho must be either None or a strictly positive number.")

        num_groups = len(set(groups_num))
        eps0 = self.eps / 2.0

        kargs = {
            "eps0": eps0,
            "rho": self.rho,
            "gamma": self.gamma,
            "sgd_steps": self.sgd_steps,
            "batch_size": self.batch_size,
        }

        self.lambdas = np.zeros((num_groups,))
        self.mus = np.zeros((num_groups,))
        logger = RTLogger(total_iterations=num_groups, verbose=self.verbose)
        for k in range(num_groups):
            group_k = np.where(groups_num == k)
            optimizer = SGDDebiaser(y=y_score[group_k], **kargs)

            # Step 1: SGD
            lambda_final = 0
            mu_final = 0
            for _ in range(self.sgd_steps):
                optimizer.update_step()
                lambda_final += optimizer.lambda_k / self.sgd_steps
                mu_final += optimizer.mu_k / self.sgd_steps

            # Step 2: Full gradient descent
            optimizer = FullGDDebiaser(
                y=y_score[group_k], lambda_0=lambda_final, mu_0=mu_final, **kargs
            )
            for _ in range(self.full_gradient_epochs):
                optimizer.update_step()

            self.lambdas[k] = optimizer.lambda_k
            self.mus[k] = optimizer.mu_k

            logger.update(iteration=k + 1)

    def predict(self, y_score, p_attr):
        """
        Debiases the predictions.

        Description
        -----------

        Given the original scores y, post-process them according to the learned
        model such that the predictions satisfy the desired fairness criteria.

        Parameters
        ----------
          y_orig: np.ndarray
            Predicted probability Vector

          groups_num: np.ndarray
            Group membership vector from group 0 to group (num_classes-1).

        Returns
        -------
          y_score: is the probability of predicting the positive
            class for the instance i.
        """
        num_examples = len(y_score)
        gamma = self.gamma
        lambdas = self.lambdas
        mus = self.mus

        new_y_score = np.zeros((num_examples,))

        for i, k in enumerate(p_attr):
            if y_score[i] < (lambdas[k] - mus[k]):
                new_y_score[i] = 0
            elif y_score[i] < (lambdas[k] - mus[k]) + gamma:
                new_y_score[i] = (1.0 / gamma) * (y_score[i] - (lambdas[k] - mus[k]))
            else:
                new_y_score[i] = 1.0

        return new_y_score



import torch
from torch.utils.data import Dataset


class ADDataset(Dataset):
    """
    Dataset class to train adversarial debiasing pytorch model.
    """

    def __init__(self, X, y, groups_num, device):
        self.X = X
        self.y = y
        self.groups_num = groups_num
        self.device = device

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        X = torch.tensor(self.X[idx]).type(torch.FloatTensor).to(self.device)
        y = torch.tensor(self.y[idx, None]).type(torch.FloatTensor).to(self.device)
        group = (
            torch.tensor(self.groups_num[idx, None])
            .type(torch.FloatTensor)
            .to(self.device)
        )
        return (X, y), (y, group)



from calendar import c
from typing import Optional

import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm import BinaryBalancerAlgorithm


class LPDebiaserBinary(BMPost):
    """
    Linear Programmin Debiaser is a postprocessing algorithms designed to debias pretrained classifiers.
    The algorithm use constraints such as Equalized Odds and Equalized Opportunity.
    References:
        Hardt, Moritz, Eric Price, and Nati Srebro. "Equality of opportunity in supervised learning."
        Advances in neural information processing systems 29 (2016).
    """

    CONSTRAINT = ["EqualizedOdds", "EqualizedOpportunity"]
    OBJ_FUN = ["macro", "micro"]

    def __init__(
        self,
        constraint: Optional["CONSTRAINT"] = "EqualizedOdds",
    ):
        """
        Parameters
        ----------
        constraint : str
            Strategy used to evalute the cost function  The available contraints  are:
            [
                "EqualizedOdds",
                "EqualizedOpportunity"
            ]
        """
        self.constraint = constraint
        self.sens_groups = SensitiveGroups()

    def fit(
        self,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        y_pred: Optional[np.ndarray] = None,
        y_proba: Optional[np.ndarray] = None,
    ):
        """

        Description
        ----------
        Compute parameters for Linear Programming Debiaser.
        For binary classification y_pred or y_proba can be used.
        For Multiclass classification only y_pred must be used.
        Parameters
        ----------
        y_true : array-like
            Target vector
        y_pred : array-like
            Predicted label vector (num_examples,).
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        Self
        """
        params = self._load_data(
            y_true=y_true,
            y_pred=y_pred,
            y_proba=y_proba,
            group_a=group_a,
            group_b=group_b,
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_true = params["y_true"]
        y_pred = params.get("y_pred", None)
        y_proba = params.get("y_proba", None)

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self.sens_groups.fit_transform(
            sensitive_features, convert_numeric=True
        )

        constraints_catalog, objective_catalog = self._get_catalogs()

        constraint = constraints_catalog[self.constraint]()

        objective = objective_catalog["losses"]()

        self.algorithm = BinaryBalancerAlgorithm(
            constraint=constraint, objective=objective
        )

        self.algorithm.fit(y_true=y_true, y_pred=y_pred, y_proba=y_proba, p_attr=p_attr)

        return self

    def transform(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        y_pred: Optional[np.ndarray] = None,
        y_proba: Optional[np.ndarray] = None,
    ):
        """
        Apply transform function to predictions and likelihoods

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dictionnary with new predictions
        """

        params = self._load_data(
            y_pred=y_pred, y_proba=y_proba, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self.sens_groups.transform(sensitive_features, convert_numeric=True)

        y_proba = params.get("y_proba", None)
        y_pred = params.get("y_pred", None)
        new_y_pred = self.algorithm.predict(
            y_pred=y_pred, y_proba=y_proba, p_attr=p_attr
        )
        return {"y_pred": new_y_pred}

    def fit_transform(
        self,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        y_pred: Optional[np.ndarray] = None,
        y_proba: Optional[np.ndarray] = None,
    ):
        """
        Fit and transform

        Parameters
        ----------
        y_true : array-like
            Target vector
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_true=y_true,
            y_proba=y_proba,
            y_pred=y_pred,
            group_a=group_a,
            group_b=group_b,
        ).transform(y_proba=y_proba, y_pred=y_pred, group_a=group_a, group_b=group_b)

    def _get_catalogs(self):
        from .constraints import EqualizedOdds, EqualizedOpportunity, Losses

        cons_cat = {
            "EqualizedOdds": EqualizedOdds,
            "EqualizedOpportunity": EqualizedOpportunity,
        }

        obj_cat = {
            "losses": Losses,
        }

        return cons_cat, obj_cat



from abc import ABC, abstractmethod

import numpy as np


class TransformerBase:
    """
    Base implementation for unconventional transformers.
    For each pipeline execution the transformer will be linked to the estimator and custom parameters.
    """

    def has_param_handler(self):
        return hasattr(self, "params_hdl")

    def link_parameters(self, params_hdl, estimator_hdl):
        self.params_hdl = params_hdl
        self.estimator_hdl = estimator_hdl

    def get_bias_param(self, param_name, default=None):
        if param_name in self.params_hdl.bias_mitigator:
            return self.params_hdl.bias_mitigator[param_name]
        else:
            return default

    def update_bias_param(self, param_name, param_value):
        if hasattr(self, "params_hdl"):
            self.params_hdl.bias_mitigator[param_name] = param_value
        else:
            self.bias_mitigator_params[param_name] = param_value

    def get_estimator_param(self, param_name, default=None):
        if param_name in self.params_hdl.estimator:
            return self.params_hdl.estimator[param_name]
        else:
            return default

    def update_estimator_param(self, param_name, param_value):
        if hasattr(self, "params_hdl"):
            self.params_hdl.estimator[param_name] = param_value
        else:
            self.estimator_params[param_name] = param_value


class BMTransformerBase(ABC, TransformerBase):
    """
    Bias Mitigation Transformer Base
    The class implement input preprocessing for all Bias Mitigators Types.
    """

    def _get_param(self, kargs, param_name):
        if hasattr(self, param_name):
            return getattr(self, param_name)
        elif param_name in kargs:
            return kargs[param_name]
        else:
            return None

    def _to_numpy(self, kargs, value_name, ravel=True):
        value = np.array(kargs[value_name])
        return value.ravel() if ravel else value

    def _load_data_pipeline(self):

        if not hasattr(self, "params_hdl"):
            return {}

        params = {}
        bm_param_names = ["group_a", "group_b"]
        for param_name in bm_param_names:
            if param_name in self.params_hdl.bias_mitigator:
                params[param_name] = self.params_hdl.bias_mitigator[param_name]
            elif hasattr(self, param_name):
                params[param_name] = getattr(self, param_name)

        es_param_names = ["sample_weight", ("y", "y_true")]
        for param_name in es_param_names:
            if isinstance(param_name, tuple):
                es_param_name, param_name = param_name
            else:
                es_param_name = param_name

            if es_param_name in self.params_hdl.estimator:
                params[param_name] = self.params_hdl.estimator[es_param_name]

        return params

    def reformat_function(self, func):
        def wrapped_func(*args, **kargs):
            self.estimator_params = {}
            self.bias_mitigator_params = {}
            fun_varnames = func.__code__.co_varnames[1:]
            params = dict(zip(fun_varnames, args))
            kargs.update(params)
            kargs.update(self._load_data_pipeline())
            params = {v: kargs[v] for v in fun_varnames if v in kargs}
            return func(**params)

        return wrapped_func

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)

        if hasattr(obj, "fit"):
            docstring = obj.fit.__doc__
            obj.fit = obj.reformat_function(obj.fit)
            obj.fit.__doc__ = docstring

        if hasattr(obj, "transform"):
            docstring = obj.transform.__doc__
            obj.transform = obj.reformat_function(obj.transform)
            obj.transform.__doc__ = docstring

        if hasattr(obj, "fit_transform"):
            docstring = obj.fit_transform.__doc__
            obj.fit_transform = obj.reformat_function(obj.fit_transform)
            obj.fit_transform.__doc__ = docstring

        if hasattr(obj, "predict"):
            docstring = obj.predict.__doc__
            obj.predict = obj.reformat_function(obj.predict)
            obj.predict.__doc__ = docstring

        if hasattr(obj, "predict_proba"):
            docstring = obj.predict_proba.__doc__
            obj.predict_proba = obj.reformat_function(obj.predict_proba)
            obj.predict_proba.__doc__ = docstring

        obj.estimator_params = {}
        obj.bias_mitigator_params = {}
        return obj



from typing import Optional

import numpy as np
from tqdm import trange

from holisticai.utils.models.recommender._rsbase import RecommenderSystemBase
from holisticai.utils.transformers.bias import BMInprocessing as BMImp


class BlindSpotAwareMF(BMImp, RecommenderSystemBase):
    """
    A blind spot aware Matrix Factorization takes into account the blind spot
    inherent in the learning phase of the RS. The blind spot size is the
    number of item with a predicted ratings Ru,i that is smaller than a
    threshold.

    References:
        [1] Sun, Wenlong, et al. "Debiasing the human-recommender system
        feedback loop in collaborative filtering." Companion Proceedings
        of The 2019 World Wide Web Conference. 2019.
    """

    def __init__(
        self,
        K: Optional[int] = 10,
        beta: Optional[float] = 0.002,
        steps: Optional[int] = 200,
        alpha: Optional[float] = 0.002,
        lamda: Optional[float] = 0.2,
        verbose: Optional[int] = 0,
    ):
        """
        Init Blind Spot Aware Matrix Factorization

        Parameters
        ----------

        K : int
            Specifies the number of dimensions.

        beta : float
            Parameter used to update P and Q.

        steps : int
            Number of iterations.

        alpha : float
            Model parameter.

        lamda : float
            Model parameter.

        verbose : int
            If >0, will show progress percentage.
        """
        self.beta = beta
        self.steps = steps
        self.alpha = alpha
        self.lamda = lamda
        self.K = K
        self.verbose = verbose

    def fit(self, X: Optional[np.ndarray], **kargs):
        """
        Fit model

        Parameters
        ----------

        X : matrix-like
            rating matrix, 0 means non-raked cases.

        P0 : matrix-like (optional)
            Initial P matrix (numUsers, K)

        Q0 : matrix-like (optional)
            Initial P matrix (numItems, K)
        """
        numUsers, numItems = X.shape
        P0 = kargs.get("P", np.random.rand(numUsers, self.K))
        Q0 = kargs.get("Q", np.random.rand(numItems, self.K))
        P, Q = self._blindspot_awarness_matrix_factorization(X, P0, Q0)
        self.pred = np.dot(P, Q)
        self.invP = None
        return self

    def _blindspot_awarness_matrix_factorization(self, R, P, Q):
        """
        Blind Spot Awarness Matrix Factorization Algorithm
        """
        Q = Q.T
        N, M = R.shape
        W = np.ones(((N, M)))
        i_list, j_list = np.where(R > 0)
        W_s = W[i_list, j_list]
        for _ in trange(self.steps, leave=self.verbose > 0):
            for i, j, wij in zip(i_list, j_list, W_s):
                eij = R[i][j] - np.dot(P[i, :], Q[:, j])
                P[i, :] = P[i, :] + self.alpha * (
                    2 * eij * Q[:, j]
                    - self.lamda * P[i, :]
                    - self.beta * (P[i, :] - Q[:, j]) * wij
                )
                Q[:, j] = Q[:, j] + self.alpha * (
                    2 * eij * P[i, :]
                    - self.lamda * Q[:, j]
                    + self.beta * (P[i, :] - Q[:, j]) * wij
                )

        return P, Q



import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from .losses import PRBinaryCrossEntropy

EPSILON = 1.0e-10
SIGMOID_RANGE = np.log((1.0 - EPSILON) / EPSILON)


class PRLogiticRegression:
    def __init__(self, loss, initializer, fit_intercept=False):
        self.fit_intercept = fit_intercept
        self.nb_clases = 2
        self.loss = loss
        self.initializer = initializer

    def init_params(self, X, y, groups):
        # set instance variables
        X = self._preprocessing_data(X)
        self.nb_features = X.shape[1]
        self.nb_group_values = len(np.unique(groups))
        self.coef = self.initializer.initialize(
            X, y, groups, self.nb_group_values, self.nb_features
        )

    def set_params(self, coef):
        self.coef = np.reshape(coef, [self.nb_group_values, self.nb_features])

    def _preprocessing_data(self, X):
        if self.fit_intercept:
            X = np.concatenate([X, np.ones((X.shape[0], 1))], axis=1)
        return X

    def predict(self, X, groups):
        return np.argmax(self.predict_proba(X, groups), 1)

    def predict_proba(self, X, groups):
        X = self._preprocessing_data(X)
        proba = np.empty((X.shape[0], self.nb_clases))
        proba[:, 1] = self.sigmoid(X=X, groups=groups)
        proba[:, 0] = 1.0 - proba[:, 1]
        return proba

    def predict_score(self, X, groups, coef=None):
        X = self._preprocessing_data(X)
        return self.sigmoid(X=X, groups=groups, coef=coef)

    def sigmoid(self, X, groups, coef=None):
        if coef is None:
            coef = self.coef
        w = coef[groups, :]
        s = np.sum(w * X, axis=1)
        s = np.clip(s, -SIGMOID_RANGE, SIGMOID_RANGE)
        return 1.0 / (1.0 + np.exp(-s))


class PRParamInitializer:
    def __init__(self, init_type="Zero", **kargs):
        self.init_type = init_type
        if init_type.startswith("StandarLR"):
            self.C = kargs["C"]
            self.penalty = kargs["penalty"]
            self.fit_intercept = kargs["fit_intercept"]

    def initialize(self, X, y, groups, nb_group_values, nb_features):
        """set initial weight
        initialization methods are specified by `itype`
        * 0: cleared by 0
        * 1: follows standard normal distribution
        * 2: learned by standard logistic regression
        * 3: learned by standard logistic regression separately according to
            the value of sensitve feature
        Parameters
        ----------
        itype : int
            type of initialization method
        X : array, shape=(n_samples, n_features)
            feature vectors of samples
        y : array, shape=(n_samples)
            target class of samples
        s : array, shape=(n_samples)
            values of sensitive features
        """

        if self.init_type == "Zero":
            # clear by zeros
            coef = np.zeros(nb_group_values * nb_features, dtype=np.float)

        elif self.init_type == "Random":
            # at random
            coef = np.random.randn(nb_group_values * nb_features)

        elif self.init_type == "StandarLR":
            # learned by standard LR
            coef = np.empty(nb_group_values * nb_features, dtype=np.float)
            coef = coef.reshape(nb_group_values, nb_features)

            clr = LogisticRegression(
                C=self.C, penalty=self.penalty, fit_intercept=self.fit_intercept
            )
            clr.fit(X, y)

            coef[:, :] = clr.coef_

        elif self.init_type == "StandarLRbyGroup":
            # learned by standard LR
            coef = np.empty(nb_group_values * nb_features, dtype=np.float)
            coef = coef.reshape(nb_group_values, nb_features)

            for i in range(nb_group_values):
                clr = LogisticRegression(
                    C=self.C, penalty=self.penalty, fit_intercept=self.fit_intercept
                )
                clr.fit(X[groups == i, :], y[groups == i])
                coef[i, :] = clr.coef_
        else:
            raise TypeError

        return coef



# Constant values defined for grid search reduction
_UPPER_BOUND_DIFF = "upper"
_LOWER_BOUND_DIFF = "lower"
_LABEL = "y"
_EVENT = "event"
_GROUP_ID = "group_id"
_P_ATTR = "p_attr"
_ALL = "all"
_PRED = "pred"
_SIGNED = "signed"
_LOSS = "loss"
_PREDICTION = "prediction"

# Constant values defined for exponentiated gradient reductions
SHRINK_ETA = 0.8
SHRINK_REGRET = 0.8
MIN_ITER = 5
REGRET_CHECK_START_T = 5
REGRET_CHECK_INCREASE_T = 1.6
ACCURACY_MUL = 0.5
PRECISION = 1e-8



import numpy as np


def load_data():
    """
    Simulates a completed rating matrix ground truth, feedback matrix time and split test data.
    """
    num_user = 500
    num_item = 500
    epsilon = 0.5
    ratings = np.zeros((num_user, num_item))
    for u in range(num_user):
        au = np.random.normal(3.4, 1, 1)
        bu = np.random.normal(0.5, 0.5, 1)
        for i in range(num_item):
            ti = np.random.normal(0.1, 1, 1)
            eij = np.random.normal(0, 1, 1)

            a = au + bu * ti + epsilon * eij
            ratings[u][i] = max(min(round(a[0]), 5), 1)

    Time_range = 40  ## split the ratings into 40 time range
    rated_time = np.random.randint(
        1, Time_range, size=(ratings.shape[0], ratings.shape[1])
    )

    ## get the unique testing for all different debias strateies
    test_mask = (27 <= rated_time) & (rated_time <= 30)
    test_ratings = ratings * test_mask

    return ratings, rated_time, test_ratings



class SparseList(list):
    def __init__(self, default=0, data=None):
        self.default = default
        self.vals = {}
        self.size = 0

        if data:
            self.extend(data)

    def __setitem__(self, index, value):
        if self.default != value:
            self.vals[index] = value
        self.size += 1

    def __len__(self):
        return self.size

    def __getitem__(self, index):
        if index in self.vals:
            return self.vals[index]
        else:
            return self.default

    def __repr__(self):
        return "<SparseList {}>".format(self.vals)

    def append(self, val):
        if self.default != val:
            self.vals[self.size] = val
        self.size += 1

    def extend(self, iterator):
        for val in iterator:
            if self.default != val:
                self.vals[self.size] = val
            self.size += 1

    def sort(self):
        values = sorted(self.vals.values())
        self.vals = {}
        old_size = self.size
        self.size = 0

        need_to_add_default = True
        for value in values:
            if need_to_add_default and self.default < value:
                self.size += old_size - len(values)
                need_to_add_default = False

            if self.default != value:
                self.vals[self.size] = value
            self.size += 1



import sys
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.utils import check_random_state

from ..commons._conventions import *
from ._lagrangian import Lagrangian


class ExponentiatedGradientAlgorithm:
    """
    This class implements the exponentiated gradient approach to reductions.

    Reference
    ---------
    `Agarwal et al. (2018) <https://arxiv.org/abs/1803.02453>`.

    """

    def __init__(
        self,
        estimator,
        constraints,
        eps: Optional[float] = 0.01,
        max_iter: Optional[int] = 50,
        nu: Optional[float] = None,
        eta0: Optional[float] = 2.0,
        verbose: Optional[int] = 0,
        seed: Optional[int] = None,
    ):
        """
        Initialize Exponentiated Gradient Reduction Algorithm

        Parameters
        ----------
        estimator : estimator
            An estimator implementing methods :code:`fit(X, y, sample_weight)` and
            :code:`predict(X)`, where `X` is the matrix of features, `y` is the
            vector of labels (binary classification) or continuous values
            (regression), and `sample_weight` is a vector of weights.
            In binary classification labels `y` and predictions returned by
            :code:`predict(X)` are either 0 or 1.
            In regression values `y` and predictions are continuous.

        constraints : ClasificationConstraint
            The disparity constraints expressed as moments

        eps : float
            Allowed fairness constraint violation; the solution is guaranteed to
            have the error within :code:`2*best_gap` of the best error under
            constraint `eps`; the constraint violation is at most
            :code:`2*(eps+best_gap)`

        max_iter : int
            Maximum number of iterations

        nu : float
            Convergence threshold for the duality gap, corresponding to a
            conservative automatic setting based on the statistical uncertainty
            in measuring classification error

        eta_0 : float
            Initial setting of the learning rate

        verbose : int
            If >0, will show progress percentage.
        """
        self.estimator = estimator
        self.constraints = constraints
        self.eps = eps
        self.max_iter = max_iter
        self.nu = nu
        self.eta0 = eta0
        self.seed = seed
        self.monitor = Monitor(verbose=verbose)
        self.eg_helper = Helper()

    def fit(self, X, y, **kwargs):
        """Return a fair classifier under specified fairness constraints.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        y : array-like
            Target vector
        """
        B = 1 / self.eps

        lagrangian = Lagrangian(X, y, self.estimator, self.constraints, B, **kwargs)

        theta = pd.Series(0, lagrangian.constraints.index)

        def compute_default_nu(h):
            absolute_error = (h(X) - self.constraints._y_as_series).abs()
            nu = (
                ACCURACY_MUL
                * absolute_error.std()
                / np.sqrt(self.constraints.total_samples)
            )
            return nu

        eta = self.eta0 / B
        gap_LP = np.PINF
        Q_LP = None
        nu = self.nu
        last_regret_checked = REGRET_CHECK_START_T
        last_gap = np.PINF
        self.monitor.max_iter = self.max_iter
        for t in range(0, self.max_iter):

            # set lambdas for every constraint
            lambda_vec = B * np.exp(theta) / (1 + np.exp(theta).sum())
            h, h_idx = lagrangian.best_h(lambda_vec)
            nu = compute_default_nu(h) if (t == 0 and nu is None) else nu
            Q_EG, result_EG = self.eg_helper.compute_EG(
                t, h_idx, lambda_vec, lagrangian, nu
            )
            gap_EG = result_EG.gap()
            gamma = lagrangian.gammas[h_idx]

            if t > 0:
                Q_LP, result_LP = self.eg_helper.compute_LP(t, lagrangian, nu)
                gap_LP = result_LP.gap()

            # keep values from exponentiated gradient or linear programming
            self.monitor.update(gap_EG, Q_EG, gap_LP, Q_LP)

            if self.monitor.stop_condition(t, nu):
                break

            # update regret
            if t >= last_regret_checked * REGRET_CHECK_INCREASE_T:
                best_gap = min(self.monitor.gaps_EG)

                if best_gap > last_gap * SHRINK_REGRET:
                    eta *= SHRINK_ETA

                last_regret_checked = t
                last_gap = best_gap

            # update theta based on learning rate
            theta += eta * (gamma - self.constraints.bound())

            self.monitor.log_progress()

        # retain relevant result data
        report = self.monitor.make_a_summary(lagrangian)
        for param in ["_hs", "weights_"]:
            setattr(self, param, report[param])
        return self

    def predict(self, X):
        """
        Provide predictions for the given input data.

        Description
        -----------
        Predictions are randomized, i.e., repeatedly calling `predict` with
        the same feature data may yield different output. This
        non-deterministic behavior is intended and stems from the nature of
        the exponentiated gradient algorithm.

        Parameters
        ----------
        X : matrix-like
            Feature data

        Returns
        -------
        Scalar or vector
            The prediction. If `X` represents the data for a single example
            the result will be a scalar. Otherwise the result will be a vector
        """
        random_state = check_random_state(self.seed)

        if self.constraints.PROBLEM_TYPE == "classification":
            positive_probs = self.predict_proba(X)[:, 1]
            return (positive_probs >= random_state.rand(len(positive_probs))) * 1
        else:
            pred = self._forward(X)
            randomized_pred = np.zeros(pred.shape[0])
            for i in range(pred.shape[0]):
                randomized_pred[i] = random_state.choice(
                    pred.iloc[i, :], p=self.weights_
                )
            return randomized_pred

    def predict_proba(self, X):
        """
        Probability mass function for the given input data.

        Description
        -----------
        For each data point, provide the probabilities with which 0 and 1 is
        returned as a prediction.

        Parameters
        ----------
        X : matrix-like
            Feature data

        Returns
        -------
        pandas.DataFrame
            Array of tuples with the probabilities of predicting 0 and 1.
        """
        pred = self._forward(X)
        positive_probs = pred[self.weights_.index].dot(self.weights_).to_frame()
        return np.concatenate((1 - positive_probs, positive_probs), axis=1)

    def _forward(self, X):
        pred = pd.DataFrame()
        for t in range(len(self._hs)):
            if self.weights_[t] == 0:
                pred[t] = np.zeros(len(X))
            else:
                pred[t] = self._hs[t](X)
        return pred


class Helper:
    """
    A helper class to handle historical data during each iteartion.
    """

    def __init__(self):
        self.Qsum = pd.Series(dtype="float64")
        self.lambda_vecs_EG_ = pd.DataFrame()
        self.lambda_vecs_LP_ = pd.DataFrame()

    def update_Qsum(self, h_idx):
        if h_idx not in self.Qsum.index:
            self.Qsum.at[h_idx] = 0.0
        self.Qsum[h_idx] += 1.0

    def compute_EG(self, t, h_idx, lambda_vec, lagrangian, nu):
        self.lambda_vecs_EG_[t] = lambda_vec
        lambda_EG = self.lambda_vecs_EG_.mean(axis=1)
        self.update_Qsum(h_idx)
        Q_EG = self.Qsum / self.Qsum.sum()
        result_EG = lagrangian.eval_gap(Q_EG, lambda_EG, nu)
        return Q_EG, result_EG

    def compute_LP(self, t, lagrangian, nu):
        Q_LP, lambda_LP, result_LP = lagrangian.solve_linprog(nu)
        self.lambda_vecs_LP_[t] = lambda_LP
        return Q_LP, result_LP


class Monitor:
    """Monitor class used to store and create a summary report"""

    def __init__(self, verbose):
        self.gaps = []
        self.Qs = []
        self.gaps_EG = []
        self.verbose = verbose
        self.step = 0

    def update(self, gap_EG, Q_EG, gap_LP, Q_LP):
        self.gaps_EG.append(gap_EG)

        if gap_EG < gap_LP:
            self.Qs.append(Q_EG)
            self.gaps.append(gap_EG)
        else:
            self.Qs.append(Q_LP)
            self.gaps.append(gap_LP)

    def stop_condition(self, t, nu):
        return (self.gaps[t] < nu) and (t >= MIN_ITER)

    def make_a_summary(self, lagrangian):
        report = {}
        gaps_series = pd.Series(self.gaps)
        gaps_best = gaps_series[gaps_series <= gaps_series.min() + PRECISION]

        report["best_iter_"] = best_iter_ = gaps_best.index[-1]
        report["best_gap_"] = self.gaps[best_iter_]
        report["last_iter_"] = len(self.Qs) - 1
        report["weights_"] = weights_ = self.Qs[best_iter_]
        report["_hs"] = _hs = lagrangian.hs

        for h_idx in _hs.index:
            if h_idx not in weights_.index:
                report["weights_"].at[h_idx] = 0.0

        report["predictors_"] = lagrangian.predictors
        report["lambda_vecs_"] = lagrangian.lambdas

        return report

    def log_progress(self):
        self.step += 1
        if self.verbose:
            sys.stdout.write(f"\rsteps: {self.step}\tBest gap:{min(self.gaps_EG):.4f}")
            sys.stdout.flush()



import numpy as np
from sklearn.pipeline import Pipeline as SKLPipeline
from sklearn.utils.metaestimators import available_if

from holisticai.pipeline._pipeline_helper import PipelineHelper


def _fulfill_conditions(fn_name: str):
    def check(self):
        if fn_name == "predict_proba":
            return (
                hasattr(self._final_estimator, "predict_proba")
                and not self.post_estimator_transformers
            )
        if fn_name == "predict_score":
            return hasattr(self, "predict_proba") and self.post_estimator_transformers
        if fn_name == "predictions":
            return self.post_estimator_transformers

    return check


class Pipeline(SKLPipeline, PipelineHelper):
    """
    Holistic AI Pipeline

    Description
    -----------
    Holisticai pipeline wrap the sklearn pipeline to support unconventional transformers.
    Unconventional transformers (u-transformers) are transformers that doesn't follow the typically
    sklearn workflow. For example, Bias Mitigator needs update inputs, outputs, and other parameters during the
    fit and transform process. The current version of this pipeline supports only binary
    classification.
    """

    def __init__(self, steps, *, memory=None, verbose=False):
        """
        Initialize Holistic AI Pipeline

        Description
        -----------
        Preprocess the steps before pass to sklearn pipeline. The preprocessing map the bias mitigators and
        and wrap the estimator so we can share paramters during the fit ans transform function.

        Parameters
        ----------
        steps: list
            A list of transformers/u-transformers and estimator

        memory : str or object with the joblib.Memory interface, default=None
            Used to cache the fitted transformers of the pipeline. By default,
            no caching is performed. If a string is given, it is the path to
            the caching directory. Enabling caching triggers a clone of
            the transformers before fitting. Therefore, the transformer
            instance given to the pipeline cannot be inspected
            directly. Use the attribute ``named_steps`` or ``steps`` to
            inspect estimators within the pipeline. Caching the
            transformers is advantageous when fitting is time consuming.

        verbose : bool, default=False
            If True, the time elapsed while fitting each step will be printed as it
            is completed.
        """
        steps = self.preprocessing_steps(steps)
        super(Pipeline, self).__init__(steps=steps, memory=memory, verbose=verbose)

    def fit(self, X, y=None, **fit_params):
        """Fit the model.

        Fit all the transformers/u-transformers one after the other and transform the
        data and parameters. Then, fit the transformed data using the final estimator.
        Finally, fit the u-transformers for postprocessing.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self : object
            Pipeline with fitted steps.
        """
        super().fit(X, y, **fit_params)
        Xt = self._transform_without_final(X)
        self.fit_post_estimator_transformers(Xt, y)
        return self

    @available_if(_fulfill_conditions("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        """Update avaiable conditions for predict_proba"""
        return super().predict_proba(X, **predict_proba_params)

    @available_if(_fulfill_conditions("predict_score"))
    def predict_score(self, X, **predict_score_params):
        """
        Return probability vector

        Description
        -----------

        Transform the data, and the postprocessor u-transformer compute the predictions for that model.
        Only available with postprocessors bias mitigator.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        np.ndarray
            probability value for each example
        """
        return self.predictions(X, **predict_score_params)["y_score"]

    @available_if(_fulfill_conditions("predictions"))
    def predictions(self, X, **params):
        """
        Post-processor prediction

        Description
        -----------

        Transform the data, and the postprocessor u-transformer compute the predictions for that model.
        If the pipeline doesn't have postprocessors, sklearn pipeline functions are used
        for prediction.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

            .. versionadded:: 0.20

        Returns
        -------
        dict
            dictionary with postprocessor outputs
        """
        Xt = self._transform_without_final(X)
        return self._transform_post_estimator_transformers(Xt, **params)



import numpy as np


def f_mincon(etaX, S, M, L, beta, lambda_):
    S = np.array(S).astype(np.int32)
    p = np.array([np.mean(S == 0), np.mean(S == 1)])

    group_values = np.unique(S)

    assert len(group_values) == 2

    def h_fn(E, S, i):
        return 2 * p[S] * i * M * E / L - p[S] * (i**2) * (M**2) / (L**2)

    def tot_fn(H, S, i):
        return np.exp(lambda_[i + L] * (2 * S - 1) / beta + H / beta)

    Hs = []
    TOTs = []
    for i in np.arange(-L, L + 1, 1):
        H_i = h_fn(etaX, S, i)
        TOT_i = tot_fn(H_i, S, i)

        Hs.append(H_i)
        TOTs.append(TOT_i)

    TOT = sum(TOTs)
    RISs = []
    for i in np.arange(2 * L + 1):
        H_i = Hs[i]
        tmp1 = TOTs[i] / TOT
        RIS_i = tmp1 * (
            (2 * S - 1) * lambda_[i] + H_i - beta * np.log((2 * L + 1) * tmp1)
        )
        RISs.append(RIS_i)

    RIS = sum(RISs)
    ris = np.mean(RIS[S == 0]) + np.mean(RIS[S == 1])

    return ris


def f_lambda(Y, S, M, L, beta):

    from functools import partial

    from scipy.optimize import minimize

    lambda_ = 0.99 * np.ones(shape=(2 * L + 1, 1))

    fun = partial(f_mincon, Y, S, M, L, beta)

    lambda_ = minimize(
        fun, lambda_, bounds=[(0, 4 * M)] * len(lambda_)
    ).x  # , method='L-BFGS-B').x

    return lambda_



import sys

import numpy as np


def process_propensities(observed_ratings, inverse_propensities):
    numObservations = np.ma.count(observed_ratings)
    numUsers, numItems = np.shape(observed_ratings)
    scale = numUsers * numItems
    inversePropensities = None
    if inverse_propensities is None:
        inversePropensities = (
            np.ones((numUsers, numItems), dtype=np.longdouble) * scale / numObservations
        )
    else:
        inversePropensities = np.array(
            inverse_propensities, dtype=np.longdouble, copy=True
        )

    inversePropensities = np.ma.array(
        inversePropensities,
        dtype=np.longdouble,
        copy=False,
        mask=np.ma.getmask(observed_ratings),
        fill_value=0,
        hard_mask=True,
    )
    return inversePropensities


def predict_scores(
    user_vectors, item_vectors, user_biases, item_biases, global_bias, use_bias=True
):
    rawScores = np.dot(user_vectors, item_vectors.T)
    if use_bias:
        biasedScores = (
            rawScores + user_biases[:, None] + item_biases[None, :] + global_bias
        )
        return biasedScores
    else:
        return rawScores


def normalize_propensities(
    inversePropensities, normalization, scale, numItems, numUsers
):
    perUserNormalizer = np.ma.sum(inversePropensities, axis=1, dtype=np.longdouble)
    perUserNormalizer = np.ma.masked_less_equal(perUserNormalizer, 0.0, copy=False)

    perItemNormalizer = np.ma.sum(inversePropensities, axis=0, dtype=np.longdouble)
    perItemNormalizer = np.ma.masked_less_equal(perItemNormalizer, 0.0, copy=False)

    globalNormalizer = np.ma.sum(inversePropensities, dtype=np.longdouble)

    normalizedPropensities = None
    if normalization == "Vanilla":
        normalizedPropensities = inversePropensities
    elif normalization == "SelfNormalized":
        normalizedPropensities = scale * np.ma.divide(
            inversePropensities, globalNormalizer
        )
    elif normalization == "UserNormalized":
        normalizedPropensities = numItems * np.ma.divide(
            inversePropensities, perUserNormalizer[:, None]
        )
    elif normalization == "ItemNormalized":
        normalizedPropensities = numUsers * np.ma.divide(
            inversePropensities, perItemNormalizer[None, :]
        )
    else:
        print("MF.GENERATE_MATRIX: [ERR]\t Normalization not supported:", normalization)
        sys.exit(0)
    return normalizedPropensities


def get_bias_configuration(bias_mode):
    useBias = None
    regularizeBias = None
    if bias_mode == "None":
        useBias = False
        regularizeBias = False
    elif bias_mode == "Regularized":
        useBias = True
        regularizeBias = True
    elif bias_mode == "Free":
        useBias = True
        regularizeBias = False
    else:
        print("MF.GENERATE_MATRIX: [ERR]\t Bias mode not supported:", bias_mode)
        sys.exit(0)
    return useBias, regularizeBias


def get_metric_mode(mode):
    metricMode = None
    if mode == "mse":
        metricMode = 1
    elif mode == "mae":
        metricMode = 2
    else:
        print("MF.GENERATE_MATRIX: [ERR]\t Metric not supported:", mode)
        sys.exit(0)
    return metricMode


def get_loss(delta, metricMode):
    if metricMode == 1:
        loss = np.square(delta)
    elif metricMode == 2:
        loss = np.abs(delta)
    else:
        sys.exit(0)
    return loss


def get_gradient_multiplier(metricMode, normalizedPropensities, delta):
    gradientMultiplier = None
    if metricMode == 1:
        gradientMultiplier = np.multiply(normalizedPropensities, 2 * delta)
    elif metricMode == 2:
        gradientMultiplier = np.zeros(np.shape(delta), dtype=np.int)
        gradientMultiplier[delta > 0] = 1
        gradientMultiplier[delta < 0] = -1
        gradientMultiplier = np.multiply(normalizedPropensities, gradientMultiplier)
    else:
        sys.exit(0)
    return gradientMultiplier


class ParameterSerializer:
    def __init__(self, numUsers, numItems, num_dimensions):
        self.numUsers = numUsers
        self.numItems = numItems
        self.num_dimensions = num_dimensions

    def serialize(
        self, user_vectors, item_vectors, user_biases, item_biases, global_bias
    ):
        allUserParams = np.concatenate((user_vectors, user_biases[:, None]), axis=1)
        allItemParams = np.concatenate((item_vectors, item_biases[:, None]), axis=1)

        allParams = np.concatenate((allUserParams, allItemParams), axis=0)
        paramVector = np.reshape(
            allParams, (self.numUsers + self.numItems) * (self.num_dimensions + 1)
        )
        paramVector = np.concatenate((paramVector, [global_bias]))
        return paramVector.astype(np.float)

    def deserialize(self, paramVector):
        globalBias = paramVector[-1]
        remainingParams = paramVector[:-1]
        allParams = np.reshape(
            remainingParams, (self.numUsers + self.numItems, self.num_dimensions + 1)
        )
        allUserParams = allParams[0 : self.numUsers, :]
        allItemParams = allParams[self.numUsers :, :]

        userVectors = (allUserParams[:, 0:-1]).astype(np.longdouble)
        userBiases = (allUserParams[:, -1]).astype(np.longdouble)

        itemVectors = (allItemParams[:, 0:-1]).astype(np.longdouble)
        itemBiases = (allItemParams[:, -1]).astype(np.longdouble)
        return userVectors, itemVectors, userBiases, itemBiases, globalBias


class Problem:
    def __init__(self, observed_ratings, normalized_propensities, config):
        self.observed_ratings = observed_ratings
        self.normalized_propensities = normalized_propensities
        self.serializer = config["serializer"]
        self.config = config

    def evaluate(self, paramVector):
        (
            userVectors,
            itemVectors,
            userBiases,
            itemBiases,
            globalBias,
        ) = self.serializer.deserialize(paramVector)
        biasedScores = predict_scores(
            userVectors,
            itemVectors,
            userBiases,
            itemBiases,
            globalBias,
            self.config["useBias"],
        )

        delta = np.subtract(biasedScores, self.observed_ratings)
        loss = get_loss(delta, self.config["metricMode"])

        weightedLoss = loss * self.normalized_propensities
        objective = np.sum(weightedLoss)

        gradientMultiplier = get_gradient_multiplier(
            self.config["metricMode"], self.normalized_propensities, delta
        )

        userVGradient = np.dot(gradientMultiplier, itemVectors)
        itemVGradient = np.dot(gradientMultiplier.T, userVectors)

        userBGradient = None
        itemBGradient = None
        globalBGradient = None
        if self.config["useBias"]:
            userBGradient = np.sum(gradientMultiplier, axis=1)
            itemBGradient = np.sum(gradientMultiplier, axis=0)
            globalBGradient = np.sum(gradientMultiplier)
        else:
            userBGradient = np.zeros_like(userBiases)
            itemBGradient = np.zeros_like(itemBiases)
            globalBGradient = 0.0

        if self.config["l2_regularization"] > 0:
            scaledPenalty = (
                1.0
                * self.config["l2_regularization"]
                * self.config["scale"]
                / (self.config["numUsers"] + self.config["numItems"])
            )
            if self.config["regularizeBias"]:
                scaledPenalty /= self.config["num_dimensions"] + 1
            else:
                scaledPenalty /= self.config["num_dimensions"]

            userVGradient += 2 * scaledPenalty * userVectors
            itemVGradient += 2 * scaledPenalty * itemVectors

            objective += scaledPenalty * np.sum(np.square(userVectors))
            objective += scaledPenalty * np.sum(np.square(itemVectors))

            if self.config["regularizeBias"]:
                userBGradient += 2 * scaledPenalty * userBiases
                itemBGradient += 2 * scaledPenalty * itemBiases
                globalBGradient += 2 * scaledPenalty * globalBias
                objective += scaledPenalty * np.sum(np.square(userBiases))
                objective += scaledPenalty * np.sum(np.square(itemBiases))
                objective += scaledPenalty * globalBias * globalBias

        gradient = self.serializer.serialize(
            userVGradient, itemVGradient, userBGradient, itemBGradient, globalBGradient
        )
        return objective, gradient


def init_random_parameters(config, start_vec=None):
    userVectorsInit = None
    itemVectorsInit = None
    userBiasesInit = None
    itemBiasesInit = None
    globalBiasInit = None
    if start_vec is None:
        userVectorsInit = np.random.standard_normal(
            (config["numUsers"], config["num_dimensions"])
        )
        itemVectorsInit = np.random.standard_normal(
            (config["numItems"], config["num_dimensions"])
        )
        userBiasesInit = np.zeros(config["numUsers"], dtype=np.float)
        itemBiasesInit = np.zeros(config["numItems"], dtype=np.float)
        globalBiasInit = 0
    else:
        userVectorsInit = start_vec[0]
        itemVectorsInit = start_vec[1]
        userBiasesInit = start_vec[2]
        itemBiasesInit = start_vec[3]
        globalBiasInit = start_vec[4]

    return (
        userVectorsInit,
        itemVectorsInit,
        userBiasesInit,
        itemBiasesInit,
        globalBiasInit,
    )
